import logging
import os
import configparser
import sys
import argparse
import glob
import importlib
import json

from typing import Dict
from typing import List
from typing import Union
from typing import Iterable

MODULE_PATH = os.path.abspath(__file__).rsplit(os.path.sep, 4)[0]
sys.path.insert(0, MODULE_PATH)

from polygenic.version import __version__
from polygenic.lib.data_access.data_accessor import VcfAccessor

from polygenic.lib.output import create_res_representation_for_model
from polygenic.lib.data_access.allele_frequency_accessor import AlleleFrequencyAccessor
from polygenic.lib.data_access.data_accessor import VcfAccessor
from polygenic.seqql.score import PolygenicRiskScore
from polygenic.seqql.score import Data
from polygenic.lib.data_access.dto import ModelDescriptionInfo


logger = logging.getLogger('polygenic')

class Sequery(object):

    def __init__(self, testConfigFile = os.path.dirname(__file__) + "/test_module_config.json", configFile = os.path.dirname(__file__) + "/wdltest.cfg", index = -1):
        self.logger = logging.getLogger(__name__)
        self.logger.info('Reading config')
        config = self.getConfig(configFile)
        testConfig = TestConfiguration(testConfigFile).getConfiguration()

        # Preparing cromwell
        self.logger.info('Preparing cromwell handler')
        self.cromwell = CromwellHandler(config)
        self.testRunner = TestRunner(testConfig, self.cromwell, index = index)

    def getConfig(self, configFile = os.path.dirname(__file__) + "/wdltest.cfg"):
        config = configparser.ConfigParser()
        config.read(configFile)
        return config

    def run(self):
        exitCode = self.testRunner.run()
        self.cromwell.stop()
        return exitCode

    def localrun(self):
        exitCode = self.testRunner.run()
        self.cromwell.stop()
        return exitCode

    def stop(self):
        self.cromwell.stop()

def expand_path(path: str) -> str:
    return os.path.abspath(os.path.expanduser(path)) if path else ''

def load_models_when_mapping_absent(directories:Iterable[str], population:str, logger) -> Dict[str, ModelDescriptionInfo]:
    model_infos = {}
    for directory in directories:
        logger.info(f"Discovering models in {directory}")
        for model_path_ in glob.glob(os.path.join(directory, '*_{}_model.py'.format(population))):
            model_path = expand_path(model_path_)
            model_fname = os.path.basename(model_path)
            description_path = model_path.replace('.py', '.json')
            description = [description_path] if os.path.exists(description_path) else []
            plot_data_path_str = model_path.replace('_traits/', '_data/').replace('_model.py', '_data.json')
            plot_data_path = plot_data_path_str if os.path.exists(plot_data_path_str) else None
            model_infos[model_path] = ModelDescriptionInfo(
                model_fname = model_fname,
                model_path = model_path,
                desc_paths = description,
                plot_data_path = plot_data_path
            )
    return model_infos

def process_model(model_description_info:ModelDescriptionInfo, vcf_accessor:VcfAccessor, allele_freq_accessor:VcfAccessor, pop:str, sample_name:str):
    package_str = os.path.dirname(model_description_info.model_path)
    if package_str not in sys.path:
        sys.path.insert(0, package_str)
    module = importlib.import_module(model_description_info.model_fname.split('.')[0])
    pop_short = pop.replace('AF', '').lstrip('_')
    #if pop_short != module.trait_was_prepared_for_population:
    #    raise ImproperPopulationForModelError(f"You requested data for population {pop_short} while the model was prepared for {module.trait_was_prepared_for_population}")
    model: PolygenicRiskScore = module.model
    data = Data(vcf_accessor, allele_freq_accessor, sample_name, pop, model)
    return data.compute_model()

def parse_model(path: str):
    model_path = expand_path(path)
    model_fname = os.path.basename(model_path)
    description_path = model_path.replace('.py', '.json')
    description = [description_path] if os.path.exists(description_path) else []
    #plot_data_path_str = model_path.replace('_traits/', '_data/').replace('_model.py', '_data.json')
    #plot_data_path = plot_data_path_str if os.path.exists(plot_data_path_str) else None
    model_info = ModelDescriptionInfo(
        model_fname = model_fname,
        model_path = model_path,
        desc_paths = description,
        plot_data_path = None
    )
    return model_info

def main(args = sys.argv[1:]):
    parser = argparse.ArgumentParser(description='')  # todo dodać opis
    parser.add_argument('-i', '--vcf', required=True, help='Vcf file with genotypes')
    parser.add_argument('-l', '--log-file', type=str, default='/var/logs/polygenic.log')
    parser.add_argument('-s', '--sample-name', type=str, help='Sample name to calculate')
    parser.add_argument('-o', '--output-directory', type=str, default="", help='Directory for result jsons.')
    parser.add_argument('-n', '--output-name-appendix', type=str, default="", help='Output file name appendix.')
    parser.add_argument('-p', '--population', type=str, default='nfe',
                        choices=['', 'nfe', 'eas', 'afr', 'amr', 'asj', 'fin', 'oth'],
                        help='''Population code:
        empty - use average allele frequency in all population,
        'nfe' - Non-Finnish European ancestry,
        'eas' - East Asian ancestry,
        'afr' - African-American/African ancestry,
        'amr' - Latino ancestry,
        'asj' - Ashkenazi Jewish ancestry, 
        'fin' - Finnish ancestry,
        'oth' - Other ancestry''')
    #parser.add_argument('--traits_dirs', nargs='+', type=str, default=[],
    #                    help='Directories containing models from the inside of this repo')
    # parser.add_argument('--actionable_dir', type=str, default='',
    #                     help='Directory containing "actionable" models')  # default='vitalleo_actionable'
    #parser.add_argument('--models_path', type=str, default='', help="Path to a directory containing models and corresponding_descriptions")
    parser.add_argument('-m', '--model', action='append', help="Path to model")
    #parser.add_argument('--mapping_json', type=str, default='', help="A file containing mapping between models and descriptions")
    parser.add_argument('--af', type=str, default='',
                        help="A file containing allele freq data")
    parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + __version__)

    parsed_args = parser.parse_args(args)

    ### setup logging ###
    log_directory = os.path.dirname(os.path.abspath(os.path.expanduser(parsed_args.log_file)))
    if log_directory:
        try:
            os.makedirs(log_directory)
        except OSError:
            pass
    logger.setLevel(logging.DEBUG)
    logging_file_handler = logging.FileHandler(parsed_args.log_file)
    logging_file_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    logging_file_handler.setFormatter(formatter)
    logger.addHandler(logging_file_handler)

    ###
    out_dir = expand_path(parsed_args.output_directory)

    population = 'AF' if not parsed_args.population else 'AF_' + parsed_args.population

    models_info = {}
    for model in parsed_args.model:
        model_info = parse_model(model)
        models_info[model_info.model_path] = model_info
    #directory = str(os.path.join(parsed_args.models_path, '*_{}_model.py'.format(parsed_args.population)))
    #for model in glob.glob(directory):
    #    model_info = parse_model(model)
    #    models_info[model_info.model_path] = model_info
    
    if not models_info:
        raise RuntimeError("No models loaded. Exiting.")

    vcf_accessor = VcfAccessor(expand_path(parsed_args.vcf))

    if parsed_args.af == "":
        allele_accessor = None
    else:    
        allele_accessor = VcfAccessor(expand_path(parsed_args.af))
    sample_names = vcf_accessor.sample_names

    if "sample_name" in parsed_args:
        sample_names = [parsed_args.sample_name]

    for sample_name in sample_names:
        results_representations = {}
        for model_path, model_desc_info in models_info.items():
            res = process_model(model_desc_info, vcf_accessor, allele_accessor, population, sample_name)
            results_representations[model_path] = create_res_representation_for_model(res, model_desc_info, parsed_args.population)
        appendix = parsed_args.output_name_appendix
        if appendix != "": 
            appendix = "-" + appendix

        with open(os.path.join(out_dir, f'{sample_name}{appendix}.sample.json'), 'w') as f:
            json.dump(results_representations, f, indent=4)


if __name__ == '__main__':
    main(sys.argv[1:])



import logging
import re
import subprocess
from typing import List
from typing import Dict
from polygenic.lib.data_access.dto import SnpDataManySamples


SEARCH_CMD_PATTERN = 'bcftools view {} {}:{}'
GENOTYPE_STRING_RE = re.compile('\\sGT(:[A-Z]{2,3})*(.+)')

logger = logging.getLogger('description_language.' + __name__)


def extract_line_containing_rs_id(file_path: str, rsid_end: str, id_: str) -> str:
    """
    :param file_path:
    :param rsid_end: 3 last characters in the rsid
    :param id_: rsid without "rs"
    :return:
    """
    p = subprocess.Popen(SEARCH_CMD_PATTERN.format(file_path, rsid_end, id_), stdout=subprocess.PIPE, shell=True)
    output, _ = p.communicate()
    return output.decode("utf-8").strip().rsplit('\n', 1)[-1]

def get_genotypes(vcf_line: str, sample_names: List[str], genotype_string_re=GENOTYPE_STRING_RE) -> SnpDataManySamples:
    splitted = vcf_line.split(None)
    rsid = splitted[2]
    ref = splitted[3]
    alts = splitted[4].split(',')

    #print("SPL: " + str(splitted) + " rsid: " + str(rsid) + " ref: " + str(ref)) 
    #print("SPL-1: " + str(splitted[9:])) 
    #print("SEARCH: " + str(genotype_string_re.search(splitted[8:])))
    #print("GROUP: " + str(genotype_string_re.search(splitted[8:]).group(0)))

    #genotype_string_all_samples = genotype_string_re.search(splitted[-1]).group(2).strip()
    genotype_string_splitted = splitted[9:]#genotype_string_all_samples.split()
    assert len(genotype_string_splitted) == len(sample_names)
    genotype_strings = {name: genotype_string.split(':')[0] for name, genotype_string in
                        zip(sample_names, genotype_string_splitted)}
    return SnpDataManySamples(ref, alts, {sample_name: genotype_string_to_letters(ref, alts, genotype_string, rsid) for sample_name, genotype_string
            in genotype_strings.items()})


def genotype_string_to_letters(ref: str, alts: List[str], genotype_string: str, rsid: str) -> List[str]:
    # genotype_string - sth line 1|0
    if '|' in genotype_string:
        indices = genotype_string.split('|')
    else:
        indices = genotype_string.split('/')
    possible_genotypes = [ref] + alts
    genotype = []
    for index in indices:
        try:
            genotype.append(possible_genotypes[int(index)])
        except ValueError:
            logger.debug('Invalid genotype for {}: {}'.format(rsid, index))
    return genotype


if __name__ == '__main__':
    # print(get_sample_names('~/PycharmProjects/imputing/example_data/imputed_genotype'))
    # concatenate_vcfs('~/PycharmProjects/imputing/example_data/imputed_genotype', '/home/wojtek/logs/all.gz')
    # make_index('/home/wojtek/logs/all.gz')
    # print(os.listdir('/home/wojtek/logs/'))
    sample_names = get_sample_names('~/PycharmProjects/imputing/example_data/imputed_genotype')
    extracted = extract_line_containing_rs_id('/home/wojtek/logs/all.gz', 'rs568182971')
    print(extracted)
    print(get_genotypes(extracted, sample_names, 'rs568182971'))



#!/usr/bin/env python3
#
# -----------------------------------------------------------------------------
# Copyright (c) 2019, Battelle National Biodefense Institute.
#
# This file is part of rsidx (https://github.com/bioforensics/rsidx)
# and is licensed under the BSD license: see LICENSE.txt.
# -----------------------------------------------------------------------------

import os
import rsidx
import sqlite3
import sys
import re

def parse_vcf(vcfstream, updateint=1e6):
    threshold = updateint
    for n, line in enumerate(vcfstream):
        if line.startswith('#'):
            continue
        chromstr, posstr, rsids, *values = line.split('\t')
        for rsid in rsids.split(";"):
            rsid = re.search(r'rs[0-9]+', rsid)
            if not rsid is None:
                yield int(rsid.group()[2:]), chromstr, int(posstr)
        if n >= threshold:
            threshold += updateint
            if threshold == updateint * 10:
                updateint = threshold
            print('[rsidx::index] processed', n, 'variants', file=sys.stderr)


def index(dbconn, vcffh, cache_size=None, mmap_size=None, logint=1e6):
    c = dbconn.cursor()
    c.execute(
        'CREATE TABLE rsid_to_coord ('
        'rsid INTEGER PRIMARY KEY, '
        'chrom TEXT NULL DEFAULT NULL, '
        'coord INTEGER NOT NULL DEFAULT 0)'
    )
    if cache_size:
        c.execute('PRAGMA cache_size = -{:d}'.format(cache_size))
    if mmap_size:
        c.execute('PRAGMA mmap_size = {:d}'.format(mmap_size))  # bytes
    dbconn.commit()

    vcfstream = parse_vcf(vcffh, updateint=logint)
    c.executemany('INSERT OR IGNORE INTO rsid_to_coord VALUES (?,?,?)', vcfstream)
    dbconn.commit()


def main(args):
    if os.path.exists(args.idx):
        message = 'WARNING: index file "{:s}" exists'.format(args.idx)
        if args.force:
            message += ', overwriting'
            try:
                os.unlink(args.idx)
            except FileNotFoundError:  # prevent exploits  # pragma: no cover
                pass
        else:
            message += ', stubbornly refusing to proceed'
        print('[rsidx]', message, file=sys.stderr)
        if not args.force:
            raise SystemExit
    with rsidx.open(args.vcf, 'r') as vcffh:
        with sqlite3.connect(args.idx) as dbconn:
            index(dbconn, vcffh, cache_size=args.cache_size,
                  mmap_size=args.mmap_size)



#!/usr/bin/env python3
#
# -----------------------------------------------------------------------------
# Copyright (c) 2019, Battelle National Biodefense Institute.
#
# This file is part of rsidx (https://github.com/bioforensics/rsidx)
# and is licensed under the BSD license: see LICENSE.txt.
# -----------------------------------------------------------------------------

import builtins
from contextlib import contextmanager
from gzip import open as gzopen
import sys

from rsidx import index
from rsidx import search
from rsidx import __main__
from rsidx import cli

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


@contextmanager
def open(filename, mode):
    if mode not in ('r', 'w'):
        raise ValueError('invalid mode "{}"'.format(mode))
    if filename in ['-', None]:
        filehandle = sys.stdin if mode == 'r' else sys.stdout
        yield filehandle
    else:
        openfunc = builtins.open
        if filename.endswith('.gz'):
            openfunc = gzopen
            mode += 't'
        with openfunc(filename, mode) as filehandle:
            yield filehandle



import copy
from typing import Dict
from typing import Union
from typing import List
from typing import Sequence
from typing import Any


def vitalleo_model_output_single_gene(module) -> Dict[str, str]:
    return {
        'how_your_result_was_calculated': module.how_your_result_was_calculated
    }


def vitalleo_model_output_polygenic(module) -> Dict[str, Union[str, Sequence[str]]]:
    return {
        'genes': module.genes,
        'information_text': [module.how_your_result_was_calculated, module.science_behind_your_result]
    }


def vitalleo_snip(variants, patient_genotypes: Dict[str, List[str]],
                  report_numeric_gene_impact: bool = True):
    ret_variant_list = []
    for variant in variants:
        rsid = variant['rsid']
        patient_genotype = sorted(patient_genotypes[rsid])
        valid_patient_genotype = validate_snip_actionable(variant["genotypes"], patient_genotype, rsid)
        new_variant = copy.deepcopy(variant)
        new_variant["patient_genotype"] = valid_patient_genotype
        if report_numeric_gene_impact:
            new_variant['gene_impact'] = valid_patient_genotype.count(new_variant['risk_allele'])
        ret_variant_list.append(new_variant)
    return ret_variant_list


def validate_snip_actionable(genotypes: Dict[str, Any], patient_genotype: List[str], rsid: str):
    assert len(patient_genotype) in (1, 2)
    patient_genotype_set = set(patient_genotype)
    proper_genotypes = [set(genotype["genotype"]) for genotype in genotypes]
    if not patient_genotype_set in proper_genotypes:
        raise RuntimeError('Improper genotype: {} for rsid: {}. Proper genotypes are: {}'.format(
            patient_genotype, rsid, proper_genotypes))
    return patient_genotype if len(patient_genotype) == 2 else patient_genotype * 2


TEST_TYPE_TO_CALLABLE = {
    "Single SNP analysis": vitalleo_model_output_single_gene,
    "Polygenic Risk Score": vitalleo_model_output_polygenic
}


def vitalleo_model_output(category: str, module) -> Dict[str, Union[str, Sequence]]:
    ret_dict = {
        'trait': module.trait,
        'test_type': module.test_type,
        'about': module.about,
        'other_factors_statement': module.other_factors_statement,
        'other_factors': module.other_factors,
        'what_your_result_means': module.what_your_result_means[category],
        'result_statement': module.result_statement[category],
        'result': category,
        'what_you_can_do_statement': module.what_you_can_do_statement[category],
        'what_you_can_do': module.what_you_can_do[category],
        'references': module.references
    }
    ret_dict.update(TEST_TYPE_TO_CALLABLE[module.test_type](module))
    return ret_dict


def numeric_output(output: float, _) -> Dict[str, float]:
    return {
        'result': output
    }

def mobigen_output() -> Dict[str, float]:pass


import argparse
import logging

import sys
import os
import urllib.request

import configparser
import tabix

# utils
## simulate
from polygenic.lib.data_access.data_accessor import VcfAccessor
from polygenic.lib.data_access.data_accessor import DataNotPresentError

# biobankuk-index
import gzip
import io

# biobankuk-get
import progressbar
import os.path

# bioabnkuk-build-model
## clumping
import subprocess
import re
## simlating
import random
import statistics
## saving
import json

logger = logging.getLogger('polygenicmaker')
config = configparser.ConfigParser()
config.read(os.path.dirname(__file__) + "/../polygenic/polygenic.cfg")

#########################
### utility: download ###
#########################

def download(url: str, output_path: str, force: bool=False, progress: bool=False):
    if os.path.isfile(output_path) and not force:
        print("Index exists: " + output_path)
        return
    print("Downloading from " + url)
    response = urllib.request.urlopen(url)
    file_size = int(response.getheader('Content-length'))
    if file_size is None:
        progress = False
    if ".gz" in url:
        response_data = gzip.GzipFile(fileobj = response)
        file_size = progressbar.UnknownLength
    else:
        response_data = response
    if progress: bar = progressbar.ProgressBar(max_value = file_size).start()
    downloaded = 0
    if not os.path.exists(os.path.dirname(output_path)):
        os.makedirs(os.path.dirname(output_path))
    bytebuffer = b''
    while (bytes := response_data.read(1024)):
            bytebuffer = bytebuffer + bytes
            downloaded = downloaded + 1024
            if not file_size == progressbar.UnknownLength: downloaded = min(downloaded, file_size)
            progress and bar.update(downloaded)
    with open(output_path, 'w') as outfile:
        outfile.write(str(bytebuffer, 'utf-8'))
    progress and bar.finish()
    return

####################
### is valid path ##
####################
def is_valid_path(path: str, is_directory: bool = False):
    if is_directory:
        if not os.path.isdir(path):
            print("ERROR: " + path + " does not exists or is not directory")
            return False
    else:
        if not os.path.isfile(path):
            print("ERROR: " + path + " does not exists or is not a file")
            return False
    return True

####################
### read table  ####
####################
def read_table(file_path: str):
    table = []
    with open(file_path, 'r') as file:
        header = file.readline().rstrip().split('\t')
        while True:
            line = file.readline().rstrip().split('\t')
            if len(line) < 2:
                break
            if not len(header) == len(line):
                print("Line and header have different leangths")
            line_dict = {}
            for header_element, line_element in zip(header, line):
                line_dict[header_element] = line_element
            table.append(line_dict)
    return table

####################
###   add af     ###
####################     
def add_af(line, af_accessor:VcfAccessor, population:str = 'nfe', rsid_column_name:str = 'rsid'):
    try:
        af = af_accessor.get_af_by_pop(line['ID'], 'AF_' + population)
        line['af'] = af[line['ALT']]
    except DataNotPresentError:
        line['af'] = 0
    except Exception:
        ### TODO convert alternate alleles
        af = af_accessor.get_af_by_pop(line['ID'], 'AF_' + population)
        line['af'] = af[list(af.keys())[1]]
    return line

####################
###   add rsid   ###
####################     
def add_rsid(line, tabix_source:open, population:str = 'nfe', rsid_column_name:str = 'ID'):
    if "rs" in line[rsid_column_name]:
        line['rsid'] = line[rsid_column_name]
    else:
        try:
            records = tabix_source.query(line['CHROM'], int(line['POS']) - 1, int(line['POS']))
            for record in records:
                line['rsid'] = record[2]                  
                print(str(line))
                break
        except Exception:
            line['rsid'] = line[rsid_column_name]
            
    return line


####################
###   simulate   ###
####################
def simulate_parameters(data, iterations: int = 1000, coeff_column_name: str = 'BETA'):
    random.seed(0)

    randomized_beta_list = []
    for _ in range(iterations):        
        randomized_beta_list.append(sum(map(lambda snp: randomize_beta(float(snp[coeff_column_name]), snp['af']), data)))
    minsum = sum(map(lambda snp: min(float(snp[coeff_column_name]), 0), data))
    maxsum = sum(map(lambda snp: max(float(snp[coeff_column_name]), 0), data))
    return {
        'mean': statistics.mean(randomized_beta_list), 
        'sd': statistics.stdev(randomized_beta_list),
        'min': minsum,
        'max': maxsum
    }

###################
### write model ###
###################

def write_model(data, description, destination):
    
    with open(destination, 'w') as model_file:
        model_file.write("from polygenic.seqql.score import PolygenicRiskScore\n")
        model_file.write("from polygenic.seqql.score import ModelData\n")
        model_file.write("from polygenic.seqql.category import QuantitativeCategory\n")
        model_file.write("\n")
        model_file.write("model = PolygenicRiskScore(\n")
        model_file.write("categories=[\n")
        model_file.write("QuantitativeCategory(from_=" + str(description['min']) + ", to=" + str(description['mean'] - 1.645 * description['sd']) + ", category_name='Reduced'),\n")
        model_file.write("QuantitativeCategory(from_=" + str(description['mean'] - 1.645 * description['sd']) + ", to=" + str(description['mean'] + 1.645 * description['sd']) + ", category_name='Average'),\n")
        model_file.write("QuantitativeCategory(from_=" + str(description['mean'] + 1.645 * description['sd']) + ", to=" + str(description['max']) + ", category_name='Increased')\n")
        model_file.write("],\n")
        model_file.write("snips_and_coefficients={\n")
        snps = []
        for snp in data:
            snps.append("'" + snp['rsid'] + "': ModelData(effect_allele='" + snp['ALT'] + "', coeff_value=" + snp['BETA'] + ")")
        model_file.write(",\n".join(snps))
        model_file.write("},\n")
        model_file.write("model_type='beta'\n")
        model_file.write(")\n")
        model_file.write("description = " + json.dumps(description, indent=4))

    return

#####################################################################################################
###                                                                                               ###
###                                   Polygenic Score Catalogue                                   ###
###                                                                                               ###
#####################################################################################################

#######################
### pgs-index #########
#######################

def pgs_index(args):
    parser = argparse.ArgumentParser(description='polygenicmaker pgs-index downloads index of gwas results from Polgenic Score Catalogue')  # todo dodać opis
    parser.add_argument('--url', type=str, default='http://ftp.ebi.ac.uk/pub/databases/spot/pgs/metadata/pgs_all_metadata_scores.csv', help='alternative url location for index')
    parser.add_argument('--output', type=str, default='', help='output directory')
    parsed_args = parser.parse_args(args)
    output_path = os.path.abspath(os.path.expanduser(parsed_args.output)) + "/pgs_manifest.tsv"
    download(parsed_args.url, output_path, force=True)
    return

#######################
### pgs-get ###########
#######################

def pgs_get(args):
    parser = argparse.ArgumentParser(description='polygenicmaker pgs-get downloads specific gwas result from polygenic score catalogue')  # todo dodać opis
    parser.add_argument('-c', '--code', type=str, required=False, help='PGS score code. Example: PGS000814')
    parser.add_argument('-o', '--output-path', type=str, default='', help='output directory')
    parser.add_argument('-f', '--force', action='store_true', help='overwrite downloaded file')
    parsed_args = parser.parse_args(args)
    url = "http://ftp.ebi.ac.uk/pub/databases/spot/pgs/scores/" + parsed_args.code + "/ScoringFiles/" + parsed_args.code + ".txt.gz"
    download(url=url, output_path=parsed_args.output_path, force=parsed_args.force, progress=True)
    return

#####################################################################################################
###                                                                                               ###
###                                   Global Biobank Engine                                       ###
###                                                                                               ###
#####################################################################################################

#######################
### gbe-index #########
#######################

def gbe_index(args):
    print("GBEINDEX")
    parser = argparse.ArgumentParser(description='polygenicmaker gbe-index downloads index of gwas results from pan.ukbb study')  # todo dodać opis
    parser.add_argument('--url', type=str, default='https://biobankengine.stanford.edu/static/degas-risk/degas_n_977_traits.tsv', help='alternative url location for index')
    parser.add_argument('--output', type=str, default='', help='output directory')
    parsed_args = parser.parse_args(args)
    output_path = os.path.abspath(os.path.expanduser(parsed_args.output)) + "/gbe_phenotype_manifest.tsv"
    download(parsed_args.url, output_path)
    return

#######################
### gbe-get ###########
#######################

def gbe_get(args):
    parser = argparse.ArgumentParser(description='polygenicmaker biobankuk-get downloads specific gwas result from pan.ukbb study')  # todo dodać opis
    parser.add_argument('--code', type=str, required=False, help='GBE phenotype code. Example: BIN1210')
    parser.add_argument('--output', type=str, default='', help='output directory')
    parser.add_argument('--force', action='store_true', help='overwrite downloaded file')
    parsed_args = parser.parse_args(args)
    url = "https://biobankengine.stanford.edu/static/PRS_map/" + parsed_args.code + ".tsv"
    output_directory = os.path.abspath(os.path.expanduser(parsed_args.output))
    output_file_name = os.path.splitext(os.path.basename(url))[0]
    output_path = output_directory + "/" + output_file_name
    download(url=url, output_path=output_path, force=parsed_args.force, progress=True)
    return

#######################
### gbe-prepare #######
#######################

def gbe_prepare_model(args):
    parser = argparse.ArgumentParser(description='polygenicmaker biobankuk-build-model constructs polygenic score model based on p value data')  # todo dodać opis
    parser.add_argument('--data', type=str, required=True, help='path to PRS file from gbe. It can be downloaded using gbe-get')
    parser.add_argument('--output', type=str, default='', help='output directory')
    parser.add_argument('--af', type=str, required=True, help='path to allele frequency vcf. It can be downloaded with biobank-get-anno')
    parser.add_argument('--pop', type=str, default='nfe', help='population: meta, AFR, AMR, CSA, EAS, EUR, MID')
    parser.add_argument('--iterations', type=float, default=1000, help='simulation iterations for mean and sd')
    parsed_args = parser.parse_args(args)
    if not is_valid_path(parsed_args.output, is_directory=True): return
    if not is_valid_path(parsed_args.data): return
    if not is_valid_path(parsed_args.af): return
    data = read_table(parsed_args.data)
    af = VcfAccessor(parsed_args.af)
    tabix_source = tabix.open(config['urls']['hg19-rsids'])
    data = [line for line in data if "rs" in line['ID']]
    data = [add_rsid(line, tabix_source) for line in data]
    data = [add_af(line, af, population = parsed_args.pop) for line in data]
    description = simulate_parameters(data)
    model_path = parsed_args.output + "/" + os.path.basename(parsed_args.data).split('.')[0] + ".py"
    write_model(data, description, model_path)
    return


#######################
### biobankuk-index ###
#######################

def biobankuk_index(args):
    parser = argparse.ArgumentParser(description='polygenicmaker biobankuk-index downloads index of gwas results from pan.ukbb study')  # todo dodać opis
    parser.add_argument('--url', type=str, default='https://pan-ukb-us-east-1.s3.amazonaws.com/sumstats_release/phenotype_manifest.tsv.bgz', help='alternative url location for index')
    parser.add_argument('--output', type=str, default='', help='output directory')
    parsed_args = parser.parse_args(args)
    output_path = os.path.abspath(os.path.expanduser(parsed_args.output)) + "/panukbb_phenotype_manifest.tsv"
    download(parsed_args.url, output_path)
    return

#######################
### biobankuk-get #####
#######################

def biobankuk_get(args):
    parser = argparse.ArgumentParser(description='polygenicmaker biobankuk-get downloads specific gwas result from pan.ukbb study')  # todo dodać opis
    parser.add_argument('--index', type=str, default='panukbb_phenotype_manifest.tsv', help='path to phenotype_manifest.tsv index file. Can be downloaded using polygenicmaker biobankuk-index command')
    parser.add_argument('--phenocode', type=str, required=False, help='biobankUK phenotype code. Example: 30600')
    parser.add_argument('--pheno_sex', type=str, default='both_sexes', help='biobankUK pheno_sex code. Example: both_sexes')
    parser.add_argument('--coding', type=str, required=True, help='biobankUK phenotype code. Example 30600')
    parser.add_argument('--modifier', type=str, required=True, help='biobankUK phenotype code. Example 30600')
    parser.add_argument('--output', type=str, default='', help='output directory')
    parser.add_argument('--force', action='store_true', help='overwrite downloaded file')
    parsed_args = parser.parse_args(args)
    # checking index file for download url
    with open(parsed_args.index, 'r') as indexfile:
        firstline = indexfile.readline()
        phenocode_colnumber = firstline.split('\t').index("phenocode")
        aws_link_colnumber = firstline.split('\t').index("aws_link")
        while True:
            line = indexfile.readline()
            if not line:
                break
            if line.split('\t')[phenocode_colnumber] != parsed_args.phenocode:
                continue
            url = line.split('\t')[aws_link_colnumber]
            break
    # downloading
    if not url is None:
        logger.info("Downloading from " + url)
        output_directory = os.path.abspath(os.path.expanduser(parsed_args.output))
        output_file_name = os.path.splitext(os.path.basename(url))[0]
        output_path = output_directory + "/" + output_file_name
        print(parsed_args.force)
        if os.path.isfile(output_path) and parsed_args.force is False:
            print("File is laready downloaded")
            return
        logger.info("Saving to " + output_path)
        response = urllib.request.urlopen(url)
        file_size = 3.5 * int(response.getheader('Content-length'))
        decompressed_file = gzip.GzipFile(fileobj=response)
        if file_size is None:
            file_size = 7078686639
        else:
            bar = progressbar.ProgressBar(max_value = file_size).start()
            downloaded = 0
            with open(output_path, 'w') as outfile:
                while (bytes := decompressed_file.read(1024)):
                    outfile.write(str(bytes, 'utf-8'))
                    downloaded = downloaded + 1024
                    bar.update(min(downloaded, file_size))
            bar.update(file_size)
            bar.finish()
    return

#############################
### biobankuk-build-model ###
#############################

def biobankuk_build_model(args):
    parser = argparse.ArgumentParser(description='polygenicmaker biobankuk-build-model constructs polygenic score model based on p value data')  # todo dodać opis
    parser.add_argument('--data', type=str, required=True, help='path to biomarkers file from biobank uk. It can be downloaded using biobankuk-get')
    parser.add_argument('--h2', type=str, help='')
    parser.add_argument('--output', type=str, default='', help='output directory')
    parser.add_argument('--anno', type=str, required=True, help='path to annotation file. It can be downloaded with biobank-get-anno')
    parser.add_argument('--pop', type=str, default='meta', help='population: meta, AFR, AMR, CSA, EAS, EUR, MID')
    parser.add_argument('--threshold', type=float, default=1e-08, help='population: meta, AFR, AMR, CSA, EAS, EUR, MID')
    parser.add_argument('--iterations', type=float, default=1000, help='simulation iterations for mean and sd')
    parsed_args = parser.parse_args(args)
    if not os.path.isdir(parsed_args.output):
        print("ERROR: " + parsed_args.output + " does not exists or is not directory")
        return
    if not os.path.isfile(parsed_args.data):
        print("ERROR: " + parsed_args.data + " does not exists or is not a file")
        return
    if not os.path.isfile(parsed_args.anno):
        print("ERROR: " + parsed_args.anno + " does not exists or is not a file")
        return
    #filter_pval(parsed_args)
    #clump(parsed_args)
    simulation_results = simulate(parsed_args)
    description = {
        'mean': simulation_results['mean'],
        'sd': simulation_results['sd'],
        'min': simulation_results['min'],
        'max': simulation_results['max'],
        'population': parsed_args.pop
    }
    save_model(parsed_args, description)
    return

def filter_pval(args):
    output_path = args.output + "/" + os.path.basename(args.data) + ".filtered"
    with open(args.data, 'r') as data, open(args.anno, 'r') as anno, open(output_path, 'w') as output:
        data_header = data.readline().rstrip().split('\t')
        anno_header = anno.readline().rstrip().split('\t')
        output.write('\t'.join(data_header + anno_header) + "\n")
        while True:
            try:
                data_line = data.readline().rstrip().split('\t')
                anno_line = anno.readline().rstrip().split('\t')
                if float(data_line[data_header.index('pval_' + args.pop)].replace('NA','1',1)) <= args.threshold:
                    output.write('\t'.join(data_line + anno_line) + "\n")
            except:
                break
    return

def clump(args):
    filtered_path = args.output + "/" + os.path.basename(args.data) + ".filtered"
    subprocess.call("plink" + 
        " --clump " + filtered_path + 
        " --clump-p1 " + str(args.threshold) +
        " --clump-r2 0.25 " + 
        " --clump-kb 1000 " + 
        " --clump-snp-field rsid " +
        " --clump-field pval_" + args.pop +
        " --vcf results/eur.phase3.biobank.set.vcf.gz " +
        " --allow-extra-chr", 
        shell=True)
    clumped_rsids = []
    with open("plink.clumped", 'r') as plink_file:
        while(line := plink_file.readline()):
            if ' rs' in line:
                line = re.sub(' +', '\t', line).rstrip().split('\t')
                clumped_rsids.append(line[3])
    try:
        os.remove("plink.clumped")
        os.remove("plink.log")
        os.remove("plink.nosex")
    except:
        pass
    filtered_path = args.output + "/" + os.path.basename(args.data) + ".filtered"
    clumped_path = args.output + "/" + os.path.basename(args.data) + ".clumped"
    with open(filtered_path, 'r') as filtered_file, open(clumped_path, 'w') as clumped_file:
        filtered_header = filtered_file.readline().rstrip().split('\t')
        clumped_file.write('\t'.join(filtered_header) + "\n")
        while True:
            try:
                filtered_line = filtered_file.readline().rstrip().split('\t')
                if filtered_line[filtered_header.index('rsid')] in clumped_rsids:
                    clumped_file.write('\t'.join(filtered_line) + "\n")
            except:
                break
    return

def simulate(args):
    clumped_path = args.output + "/" + os.path.basename(args.data) + ".clumped"
    random.seed(0)
    simulation_data = []
    with open(clumped_path, 'r') as clumped_file:
        clumped_header = clumped_file.readline().rstrip().split('\t')
        clumped_line = clumped_header
        while True:
            clumped_line = clumped_file.readline().rstrip().split('\t')
            if len(clumped_line) < 2:
                break
            rsid = clumped_line[clumped_header.index('rsid')]
            af = float(clumped_line[clumped_header.index('af_' + args.pop)])
            beta = float(clumped_line[clumped_header.index('beta_' + args.pop)])
            simulation_data.append({'rsid': rsid, 'af': af, 'beta': beta})

    randomized_beta_list = []
    for _ in range(args.iterations):
        randomized_beta_list.append(sum(map(lambda snp: randomize_beta(snp['beta'], snp['af']), simulation_data)))
    minsum = sum(map(lambda snp: min(snp['beta'], 0), simulation_data))
    maxsum = sum(map(lambda snp: max(snp['beta'], 0), simulation_data))
    return {
        'mean': statistics.mean(randomized_beta_list), 
        'sd': statistics.stdev(randomized_beta_list),
        'min': minsum,
        'max': maxsum
        }

def randomize_beta(beta: float, af: float):
    first_allele_beta = beta if random.uniform(0, 1) < af else 0
    second_allele_beta = beta if random.uniform(0, 1) < af else 0
    return first_allele_beta + second_allele_beta

def save_model(args, description):
    model_path = args.output + "/" + os.path.basename(args.data).split('.')[0] + ".py"
    with open(model_path, 'w') as model_file:
        model_file.write("from polygenic.seqql.score import PolygenicRiskScore\n")
        model_file.write("from polygenic.seqql.score import ModelData\n")
        model_file.write("from polygenic.seqql.category import QuantitativeCategory\n")
        model_file.write("\n")
        model_file.write("model = PolygenicRiskScore(\n")
        model_file.write("categories=[\n")
        model_file.write("QuantitativeCategory(from_=" + str(description['min']) + ", to=" + str(description['mean'] - 1.645 * description['sd']) + ", category_name='Reduced'),\n")
        model_file.write("QuantitativeCategory(from_=" + str(description['mean'] - 1.645 * description['sd']) + ", to=" + str(description['mean'] + 1.645 * description['sd']) + ", category_name='Average'),\n")
        model_file.write("QuantitativeCategory(from_=" + str(description['mean'] + 1.645 * description['sd']) + ", to=" + str(description['max']) + ", category_name='Increased')\n")
        model_file.write("],\n")
        model_file.write("snips_and_coefficients={\n")
        clumped_path = args.output + "/" + os.path.basename(args.data) + ".clumped"
        snps = []
        with open(clumped_path, 'r') as clumped_file:
            clumped_header = clumped_file.readline().rstrip().split('\t')
            while True:
                clumped_line = clumped_file.readline().rstrip().split('\t')
                if len(clumped_line) < 2:
                    break
                rsid = clumped_line[clumped_header.index('rsid')]
                allele = clumped_line[clumped_header.index('alt')]
                beta = str(float(clumped_line[clumped_header.index('beta_' + args.pop)]))
                snps.append("'" + rsid + "': ModelData(effect_allele='" + allele + "', coeff_value=" + beta + ")")
        model_file.write(",\n".join(snps))
        model_file.write("},\n")
        model_file.write("model_type='beta'\n")
        model_file.write(")\n")
        model_file.write("description = " + json.dumps(description, indent=4))

    return

def main(args = sys.argv[1:]):
    try:
        if args[0] == 'biobankuk-index':
            biobankuk_index(args[1:])
        elif args[0] == 'biobankuk-get':
            biobankuk_get(args[1:])
        elif args[0] == 'biobankuk-build-model':
            biobankuk_build_model(args[1:])
        elif args[0] == 'gbe-index':
            gbe_index(args[1:])
        elif args[0] == 'gbe-get':
            gbe_get(args[1:])
        elif args[0] == 'gbe-prepare-model':
            gbe_prepare_model(args[1:])
        elif args[0] == 'pgs-index':
            pgs_index(args[1:])
        elif args[0] == 'pgs-get':
            pgs_get(args[1:])
        elif args[0] == 'pgs-prepare-model':
            pgs_prepare_model(args[1:])
        else:
            raise Exception()
    except Exception as e:
        print("ERROR " + str(e))
        print("""
        Program: polygenicmaker (downloads gwas data, clumps and build polygenic scores)
        Contact: Marcin Piechota <piechota.marcin@gmail.com>

        Usage:   polygenicmaker <command> [options]
\
        Command: 
        biobankuk-index         downloads pan biobankuk index of gwas results
        biobankuk-get           downloads gwas results for given phenocode
        biobankuk-build-model   build polygenic score based on gwas results

        """)

if __name__ == '__main__':
    main(sys.argv[1:])



import logging
import pathlib
from typing import Dict
from typing import List
from typing import Union
from polygenic.lib import mobigen_utils
from polygenic.lib.data_access.dto import SnpData
from polygenic.lib.data_access.dto import SnpDataManySamples
from polygenic.lib.data_access.vcfrecord import VcfRecord

# rsidx
import os
from gzip import open as gzopen
import rsidx
import sqlite3

logger = logging.getLogger('description_language.' + __name__)

class VcfAccessor(object):
    def __init__(self, vcf_path:str):
        super().__init__()
        self.path = vcf_path
        if not os.path.exists(self.path + '.idx.db'):
            with sqlite3.connect(self.path + '.idx.db') as dbconn, gzopen(self.path, 'rt') as vcffh:
                rsidx.index.index(dbconn, vcffh)
        self.sample_names = self.get_sample_names()
        self.__data: Dict[str, Dict[str:SnpData]] = {}  # dictionary rsid:{sample_name:ModelSnpData}

    def get_sample_names(self) -> List[str]:
        logger.info('Getting sample names')
        sample_names_for_all_files = []
        with gzopen(self.path) as vcf_file:
            for line in vcf_file:
                line = line.decode("utf-8")
                if line.find("#CHROM") != -1:
                    break
        if line.find('FORMAT') == -1:
            return None
        samples_string = line.split('FORMAT')[1].strip()
        sample_names_for_all_files.append(samples_string.split())
        assert all(sample_names == sample_names_for_all_files[-1] for sample_names in sample_names_for_all_files[:-1])
        return sample_names_for_all_files[-1]

    def __get_data_for_given_rsid(self, rsid, imputed:bool = False) -> Dict[str, SnpData]:
        line = self.__get_vcf_line_for_rsid(rsid)
        if not line:
            logger.debug(f'No line for rsid {rsid} found')
            raise DataNotPresentError
        if VcfRecord(line).is_imputed() == imputed:
            data = mobigen_utils.get_genotypes(line, self.sample_names)
            self.__data[rsid] = {sample_name: SnpData(data.ref, data.alts, genotype) for sample_name, genotype in data.genotypes.items()}
        else:
            raise DataNotPresentError
        return self.__data[rsid]

    def __get_record_for_rsid(self, rsid) -> VcfRecord:
        return VcfRecord(self.__get_vcf_line_for_rsid(rsid), self.sample_names)

    def get_af_by_pop(self, rsid:str, population_name:str) -> Dict[str, float]:
        return self.__get_record_for_rsid(rsid).get_af_by_pop(population_name)


    def get_data_for_sample(self, sample_name:str, rsid:str, imputed:bool = False) -> SnpData:
        try:
            return self.__data[rsid][sample_name]
        except KeyError:
            try:
                return self.__get_data_for_given_rsid(rsid, imputed)[sample_name]
            except DataNotPresentError:
                return None

    def __get_vcf_line_for_rsid(self, rsid:str) -> Union[None, str]:
        try:
            with sqlite3.connect(self.path + '.idx.db') as dbconn:
                for line in rsidx.search.search([rsid], dbconn, self.path):
                    return line
        except KeyError:
            print("Record " + str(rsid) + " not found")
            raise DataNotPresentError
        raise DataNotPresentError

    def get_allele_freq_from_db(rsid: str, population_name: str):
        record = self.__get_record_for_rsid(rsid)
        ref_allele = record.get_ref()
        alt_allele = record.get_alt()
        alt_allele_freq = record.get_af_by_pop(population_name)
        if not len(alt_allele) == 1:
            logger.info(
                f'{rsid} is multiallelic but only two alleles are provided. Only {ref_allele} and {alt_allele} were considered')
        return {alt_allele: alt_allele_freq, ref_allele: 1 - alt_allele_freq}

class DataNotPresentError(RuntimeError):
    pass

def path_to_fname_stem(path:str) -> str:
    return pathlib.PurePath(path).name.split('.')[0]


# import os
# import json
# from typing import Dict
# from typing import Union
# from functools import partial
# from polygenic.lib.allele_frequency_sqlite_dao import read_multiple_rs


# class AlleleFrequencyAccessor(object):

#     def __init__(self, allele_freq_json_path: str = '', allele_freq_database_path: str = ''):
#         assert not (allele_freq_json_path and allele_freq_database_path)
#         if allele_freq_json_path:
#             with open(allele_freq_json_path) as f:
#                 content = json.load(f)
#             self.reading_method = partial(get_allele_freq_from_json, allele_freq_dict=content)
#         if allele_freq_database_path:
#             assert os.path.isfile(allele_freq_database_path)
#             if not allele_freq_database_path.endswith('allele_frequency_db_gnomad_2.1.1.sqlite'):
#                 raise RuntimeError("implemented only for gnomad 2.1.1")
#             self.reading_method = partial(get_allele_freq_from_db, path=allele_freq_database_path)

#     def __call__(self, rsid: str, population: str):
#         return self.reading_method(rsid, population)


# def get_allele_freq_from_db(rsid: str, population_name: str, path: str = ''):
#     snp_data = read_multiple_rs([rsid], path=path)[0]
#     ref_allele = snp_data.ref
#     alt_allele = snp_data.alt
#     alt_allele_freq = snp_data.__dict__[population_name]
#     return {alt_allele: alt_allele_freq, ref_allele: 1 - alt_allele_freq}


# def get_allele_freq_from_json(rsid: str, population_name: str, allele_freq_dict:Dict[str, Dict[str, Union[str, float, bool]]]):
#     snp_data = allele_freq_dict[rsid]
#     ref_allele = snp_data['ref']
#     alt_allele = snp_data['alt']
#     alt_allele_freq = snp_data[population_name]
#     return {alt_allele: alt_allele_freq, ref_allele: 1 - alt_allele_freq}

# if __name__ == '__main__':
#     afa = AlleleFrequencyAccessor(
#         allele_freq_database_path='/home/wojtek/PycharmProjects/mobigen/src/main/resources/databases/allele_frequency_db_gnomad_2.1.1.sqlite')
#     print(afa('rs925966', 'AF_nfe'))
#     afa2 = AlleleFrequencyAccessor(allele_freq_json_path='/home/wojtek/PycharmProjects/mobigen/src/main/resources/allele_frequencies/gnomad.json')
#     print(afa2('rs925966', 'AF_asj'))



from typing import Dict
from typing import List
from typing import Union
from typing import NamedTuple


class SnpData(NamedTuple):
    ref:str
    alts:List[str]
    genotype:List[str]


class SnpDataManySamples(NamedTuple):
    ref: str
    alts: List[str]
    genotypes: Dict[str, List[str]] # {sample: [A, G]}


class ModelDescriptionInfo(NamedTuple):
    model_fname:str
    model_path: str
    desc_paths:List[str]
    plot_data_path:Union[str, None] = None


import numpy as np
from typing import Iterable


class QuantitativeCategory(object):
    def __init__(self, category_name, from_=-np.inf, to=np.inf, scale_from = None, scale_to = None):
        super().__init__()
        self.from_ = from_
        self.to = to
        self.name = category_name
        self.scale_from = scale_from
        self.scale_to = scale_to

    def intersects_with(self, other) -> None:
        if self.from_ == other.from_ or self.to == other.to:
            return True
        elif self.from_ < other.from_ < self.to:
            return True
        elif other.from_ < self.from_ < other.to:
            return True
        return False

    def scale_cat(self, other) -> float:
        if self.scale_from is None or self.scale_to is None:
            return other
        else:
            return self.scale_from + (self.scale_to - self.scale_from) * ((other - self.from_) / (self.to - self.from_))


    def __repr__(self):
        return f'{self.__class__.__name__}({self.name}, {self.from_}, {self.to})'


class InvalidCategoriesError(RuntimeError):
    pass


def validate_categories_names(categories: Iterable[QuantitativeCategory]):
    sorted_according_to_names = sorted(categories, key=lambda x: x.name)
    for i in range(len(sorted_according_to_names) - 1):
        if categories_have_same_names(sorted_according_to_names[i], sorted_according_to_names[i + 1]):
            raise InvalidCategoriesError(
                'Categories {} and {} have the same name'.format(sorted_according_to_names[i],
                                                                 sorted_according_to_names[i + 1]))

def categories_have_same_names(category1: QuantitativeCategory, category2: QuantitativeCategory):
    return category1.name == category2.name

def validate_categories_ranges(categories: Iterable[QuantitativeCategory]):
    for i in range(len(categories) - 1):
        for j in range(i + 1, len(categories)):
            if categories[i].intersects_with(categories[j]):
                raise InvalidCategoriesError('Categories {} and {} intersect'.format(categories[i], categories[j]))


import sqlite3
import os
import sys

module_path = os.path.abspath(__file__).rsplit(os.path.sep, 5)[0]
sys.path.insert(0, module_path)

import argparse
from collections import OrderedDict
from typing import List
from polygenic.lib.gnomad_parser import SNPdata
from polygenic.lib.gnomad_parser import get_fields
from polygenic.lib.gnomad_parser import get_data_with_unique_rsid_and_non_unique_containing_most_frequent_alt

DB_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)).rsplit('/', 2)[0], 'resources', 'databases',
                       'allele_frequency_db_gnomad_2.1.1.sqlite')#'allele_frequency_db_gnomad_3.0.sqlite')


def create_table(path: str = DB_PATH):
    if not os.path.exists(os.path.dirname(DB_PATH)):
        os.makedirs(os.path.dirname(DB_PATH))
    with sqlite3.connect(path) as conn:
        cur = conn.cursor()
        # query = '''
        #             CREATE TABLE IF NOT EXISTS allele_frequency
        #                 (
        #                     rsid VARCHAR PRIMARY KEY,
        #                     ref VARCHAR,
        #                     alt VARCHAR,
        #                     AF FLOAT,
        #                     AF_nfe FLOAT,
        #                     AF_eas FLOAT,
        #                     AF_afr FLOAT,
        #                     AF_amr FLOAT,
        #                     AF_asj FLOAT,
        #                     AF_fin FLOAT,
        #                     AF_ami FLOAT,
        #                     AF_oth FLOAT,
        #                     id_unique BOOLEAN
        #                 )
        #         '''
        query = '''
                    CREATE TABLE IF NOT EXISTS allele_frequency
                        (
                            rsid VARCHAR PRIMARY KEY,
                            ref VARCHAR,
                            alt VARCHAR,
                            AF FLOAT,
                            AF_nfe FLOAT,
                            AF_eas FLOAT,
                            AF_afr FLOAT,
                            AF_amr FLOAT,
                            AF_asj FLOAT,
                            AF_fin FLOAT,
                            AF_oth FLOAT,
                            id_unique BOOLEAN
                        )
                '''
        cur.execute(query)


def insert_many(data: List[SNPdata], path: str = DB_PATH):
    with sqlite3.connect(path) as conn:
        cur = conn.cursor()
        #query = 'INSERT INTO allele_frequency VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)'
        query = 'INSERT INTO allele_frequency VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)'
        cur.executemany(query, data)


def read_multiple_rs(rs_ids: List[str], path: str = DB_PATH) -> List[SNPdata]:
    if not os.path.exists(path):
        raise IOError('Path {} does not exist'.format(path))
    with sqlite3.connect(path) as conn:
        cur = conn.cursor()
        query = 'SELECT * FROM allele_frequency WHERE rsid IN ({})'.format(
            ', '.join('"{}"'.format(rs_id) for rs_id in rs_ids))
        cur.execute(query)
        return [SNPdata(*x) for x in cur.fetchall()]


def get_column_names(path: str = DB_PATH) -> List[str]:
    with sqlite3.connect(path) as conn:
        cur = conn.execute('select * from allele_frequency')
        return [description[0] for description in cur.description]


if __name__ == '__main__':
    #print(get_column_names())
    parser = argparse.ArgumentParser(description='')  # todo dodać opis
    parser.add_argument('data_dir', type=str, help='Directory containing gnomad data')
    parsed_args = parser.parse_args()
    create_table()
    data = []
    # for snpdata in get_fields('/home/wojtek/Pobrane/gnomad.exomes.r2.1.1.sites.Y.vcf.bgz', ['ID', 'REF', 'ALT'],
    #                           OrderedDict([('AF', float), ('AF_nfe', float), ('AF_eas', float), ('AF_afr', float),
    #                                        ('AF_amr', float), ('AF_asj', float), ('AF_fin', float), ('AF_oth', float),
    #                                        ('AF_sas', float)])):
    #     if snpdata.rsid:
    #         data.append(snpdata)
    # print('filtering')
    # data = get_data_with_unique_rsid_and_non_unique_containing_most_frequent_alt(data)
    # print(len(data))
    # print(data[-1])
    # insert_many(data)
    #raise

    for chrom in list(range(1, 23)) + ['X', 'Y']:
        print("Processing chromosome {}".format(chrom))
        file_name = 'gnomad.genomes.r3.0.sites.chr{}.vcf.bgz'.format(chrom)
        file_path = os.path.join(parsed_args.data_dir, file_name)
        data = []
        for snpdata in get_fields(file_path, ['ID', 'REF', 'ALT'],
                                  OrderedDict([('AF', float), ('AF_nfe', float), ('AF_eas', float), ('AF_afr', float),
                                               ('AF_amr', float), ('AF_asj', float), ('AF_fin', float),
                                               ('AF_ami', float), ('AF_oth', float)])):
            if snpdata.rsid and snpdata.rsid != '.':
                data.append(snpdata)
        print('filtering')
        data = get_data_with_unique_rsid_and_non_unique_containing_most_frequent_alt(data)
        print("inserting data")
        insert_many(data)
    print(read_multiple_rs(['rs59514816', 'rs1214589876']))



#!/usr/bin/env python3
#
# -----------------------------------------------------------------------------
# Copyright (c) 2019, Battelle National Biodefense Institute.
#
# This file is part of rsidx (https://github.com/bioforensics/rsidx)
# and is licensed under the BSD license: see LICENSE.txt.
# -----------------------------------------------------------------------------

import argparse
import rsidx
import sys


def main(args=None):
    """Entry point for the happer CLI.

    Isolated as a method so that the CLI can be called by other Python code
    (e.g. for testing), in which case the arguments are passed to the function.
    If no arguments are passed to the function, parse them from the command
    line.
    """
    if args is None:  # pragma: no cover
        if len(sys.argv) == 1:
            rsidx.cli.get_parser().parse_args(['-h'])
        args = rsidx.cli.get_parser().parse_args()

    versionmessage = '[rsidx] running version {}'.format(rsidx.__version__)
    print(versionmessage, file=sys.stderr)
    mainmethod = rsidx.cli.mains[args.subcmd]
    mainmethod(args)



import gzip
import re
import numpy as np
from collections import OrderedDict
from typing import List
from typing import Dict
from typing import Callable
from collections import namedtuple

##INFO=<ID=AF,Number=A,Type=Float,Description="Alternate allele frequency in samples">
##INFO=<ID=AF_nfe,Number=A,Type=Float,Description="Alternate allele frequency in samples of Non-Finnish European ancestry">
##INFO=<ID=AF_eas,Number=A,Type=Float,Description="Alternate allele frequency in samples of East Asian ancestry">
##INFO=<ID=AF_afr,Number=A,Type=Float,Description="Alternate allele frequency in samples of African-American/African ancestry">
##INFO=<ID=AF_amr,Number=A,Type=Float,Description="Alternate allele frequency in samples of Latino ancestry">
##INFO=<ID=AF_asj,Number=A,Type=Float,Description="Alternate allele frequency in samples of Ashkenazi Jewish ancestry">
##INFO=<ID=AF_fin,Number=A,Type=Float,Description="Alternate allele frequency in samples of Finnish ancestry">
##INFO=<ID=AF_oth,Number=A,Type=Float,Description="Alternate allele frequency in samples of Other ancestry">

#SNPdata = namedlist('SNPdata', 'rsid ref alt AF AF_nfe AF_eas AF_afr AF_amr AF_asj AF_fin AF_ami AF_oth is_unique',
SNPdata = namedtuple('SNPdata', ['rsid', 'ref', 'alt', 'AF', 'AF_nfe', 'AF_eas', 'AF_afr', 'AF_amr', 'AF_asj', 'AF_fin', 'AF_oth', 'is_unique'])


def get_fields(path: str, basic_fields: List[str], info_fields: Dict[str, Callable]):
    with gzip.open(path, 'rt') as f:
        for line in f:
            if not line.startswith('#'):
                CHROM, POS, ID, REF, ALT, QUAL, FILTER, info = line.split()
                d = {'CHROM':CHROM, 'POS':POS, 'ID':ID, 'REF':REF, 'ALT':ALT, 'QUAL':QUAL, 'FILTER':FILTER}
                out_dict = {x: d[x] for x in basic_fields}
                out_dict.update(get_fields_from_info_string(info, info_fields))
                fields = basic_fields + list(info_fields.keys())
                yield SNPdata(*([out_dict[field] for field in fields]))


def get_fields_from_info_string(info: str, fields: Dict[str, Callable]):
    ret_dict = {}
    for field, type_ in fields.items():
        found = re.search('{}=(.+?);'.format(field), info)
        try:
            ret_dict[field] = type_(found.group(1))
        except AttributeError:
            ret_dict[field] = None
    return ret_dict


def get_data_with_unique_rsid_and_non_unique_containing_most_frequent_alt(data: List[SNPdata]):
    data.sort(key=lambda snp_data: snp_data.rsid)
    unique_data = []
    non_unique_data = {}
    i = 0
    while i < len(data):
        j = i + 1
        while j < len(data) and data[i].rsid == data[j].rsid:
            append_data_to_list_value_of_dict(non_unique_data, data[i].rsid, data[j])
            j += 1
        if j == i + 1:
            unique_data.append(data[i])
        else:
            append_data_to_list_value_of_dict(non_unique_data, data[i].rsid, data[i])
        i = j
    for element in unique_data:
        element.is_unique = True
    for v in non_unique_data.values():
        data_containing_most_frequent_alt_allele = get_data_containing_most_frequent_alt_allele(v)
        data_containing_most_frequent_alt_allele.is_unique = False
        unique_data.append(data_containing_most_frequent_alt_allele)
    return unique_data


def append_data_to_list_value_of_dict(dictionary: Dict[str, List[SNPdata]], key: str, value_to_append: SNPdata):
    if key not in dictionary:
        dictionary[key] = [value_to_append]
    else:
        dictionary[key].append(value_to_append)


def get_data_containing_most_frequent_alt_allele(data: List[SNPdata]):
    return max(data, key=lambda snp_data: cast_nonetype_to_negative_inf(snp_data.AF))


def cast_nonetype_to_negative_inf(float_or_nonetype):
    if float_or_nonetype is None:
        return np.NINF
    return float_or_nonetype


if __name__ == '__main__':
    data = []
    for snpdata in get_fields('/home/wojtek/Pobrane/gnomad.exomes.r2.1.1.sites.Y.vcf.bgz', ['ID', 'REF', 'ALT'],
                              OrderedDict([('AF', float), ('AF_nfe', float), ('AF_eas', float), ('AF_afr', float),
                                          ('AF_amr', float), ('AF_asj', float), ('AF_fin', float), ('AF_oth', float)])):
        if snpdata.rsid and snpdata.rsid != '.':
            data.append(snpdata)
    print('filtering')
    data = get_data_with_unique_rsid_and_non_unique_containing_most_frequent_alt(data)



import math
import logging
import operator
import functools
import numpy as np
from typing import Set
from typing import Callable
from typing import Iterable
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import NamedTuple
from polygenic.lib.data_access.allele_frequency_accessor import AlleleFrequencyAccessor
from polygenic.seqql.category import validate_categories_names
from polygenic.seqql.category import validate_categories_ranges
from polygenic.seqql.category import QuantitativeCategory
from polygenic.lib.data_access.data_accessor import VcfAccessor
from polygenic.lib.data_access.dto import SnpData
from polygenic.lib.data_access.data_accessor import DataNotPresentError

logger = logging.getLogger('description_language.' + __name__)


class FunctionData(NamedTuple):
    function_to_apply_on_allele_pair: Callable
    function_to_apply_between_allele_pairs: Callable
    neutral_number: float
    function_to_compute_ranges: Callable


def beta(alleles: Iterable[str], effect_allele: str, beta_value: float) -> float:
    return alleles.count(effect_allele) * beta_value


def odds_ratio(alleles: Iterable[str], effect_allele: str, odds_ratio_value: float) -> float:
    if alleles[0] == alleles[1] == effect_allele:
        return odds_ratio_value
    elif effect_allele in alleles:
        return 1
    else:
        return 1 / odds_ratio_value


def prbs(alleles: Iterable[str], effect_allele: str, odds_ratio_value: float) -> float:
    return alleles.count(effect_allele) * math.log10(odds_ratio_value)


def sub(a, b):
    return a - b


def div_and_ln(a, b):
    return math.log(a / b)


FUNCTION_DICT = {
    'beta': FunctionData(beta, sum, 0, sub),
    'odds_ratio': FunctionData(odds_ratio, functools.partial(functools.reduce, operator.mul), 1, div_and_ln),
    'prbs': FunctionData(prbs, sum, 0, sub)
}


class ConflictingAlleleBetweenDataAndModel(RuntimeError):
    pass


class ModelData(NamedTuple):
    effect_allele: str
    coeff_value: float


class PolygenicRiskScoreResult(NamedTuple):
    initial_range_fraction: float
    imputed_range_fraction: float
    pop_freq_range_fraction: float
    missing_data_range_fraction: float
    initial_result: float
    imputed_result: float
    pop_freq_result: float
    missing_data_result: float
    result: float
    scaled_result: float
    category: str
    boundaries: Dict[str, Dict[str, float]]


class Data(object):
    def __init__(self, vcf_data_accessor: VcfAccessor,
                 allele_freq_in_population_accessor: VcfAccessor, sample_name: str, population: str, model):
        self.__vcf_data_accessor = vcf_data_accessor
        self.__allele_freq_in_population_accessor = allele_freq_in_population_accessor
        self._sample_name = sample_name
        self._population = population
        self._initial_data: Dict[str, SnpData] = {}
        self._imputed_data: Dict[str, SnpData] = {}
        self._allele_freq_in_population: Dict[str, Tuple[str]] = {}
        self._rsids_absent_in_allele_freq_data: Set[str] = set()
        self._model = model
        self._read_data()

    @property
    def sample_name(self):
        return self._sample_name

    @property
    def population(self):
        return self._population

    @property
    def initial_data(self):
        return self._initial_data

    @property
    def imputed_data(self):
        return self._imputed_data

    @property
    def allele_freq_in_population(self):
        return self._allele_freq_in_population

    @property
    def rsids_absent_in_allele_freq_data(self):
        return self._rsids_absent_in_allele_freq_data

    def _read_data(self) -> None:
        required_rsids = self._model.get_rsids()
        #print(required_rsids)
        logger.debug('Looking for rsids in initial data')
        self._initial_data = rsids_data_from_data_source(required_rsids, self.__vcf_data_accessor,
                                                         self._sample_name, False)
        self._model.validate_allele_collection(self._initial_data, 'initial data')
        rsids_not_found_in_initial_data: Set[str] = required_rsids - set(self._initial_data)
        #print(rsids_not_found_in_initial_data)
        logger.debug('Looking for rsids in initial data')
        self._imputed_data = rsids_data_from_data_source(rsids_not_found_in_initial_data, self.__vcf_data_accessor,
                                                         self._sample_name, True)
        self._model.validate_allele_collection(self._imputed_data, 'imputed data')
        rsids_not_found_in_imputed_data: Set[str] = rsids_not_found_in_initial_data - set(self._imputed_data)
        #print(rsids_not_found_in_imputed_data)
        logger.debug('Looking for rsids in allele frequency data')
        for rsid in rsids_not_found_in_imputed_data:
            #print(rsid)
            try:
                #print("====> AF " + rsid + " " + self._population)
                #print(self.__allele_freq_in_population_accessor)
                allele_frequencies: Dict[str, float] = self.__allele_freq_in_population_accessor.get_af_by_pop(rsid, self._population)
                #print(allele_frequencies)
                alleles = set(allele_frequencies)
                validate_alleles(self._model.snips_and_coefficients[rsid].effect_allele, alleles, rsid,
                                 'allele frequency data')
                self._allele_freq_in_population[rsid] = max(hardy_weinberg_diploid_frequencies(allele_frequencies).items(), key=lambda x: x[1])[0]
                
            except KeyError:
                #print("keyerror")
                logger.debug(
                    f'Either rsid: {rsid} or population: {self._population} not found in allele frequency data')
                self._rsids_absent_in_allele_freq_data.add(rsid)
                continue
            except AttributeError:
                #print("attributeerror")
                logger.debug(
                    f'Either rsid: {rsid} or population: {self._population} not found in allele frequency data')
                self._rsids_absent_in_allele_freq_data.add(rsid)
                continue
            except DataNotPresentError:
                #print("datanotpresenterror")
                logger.debug(
                    f'Either rsid: {rsid} or population: {self._population} not found in allele frequency data')
                self._rsids_absent_in_allele_freq_data.add(rsid)
                continue
            except WrongAlleleInModelError:
                logger.debug(
                    f'Rsid {rsid} has bad effect allele'
                )
                continue


    def compute_model(self) -> PolygenicRiskScoreResult:

        function_data = FUNCTION_DICT[self._model.model_type]
        result_for_initial_data = compute_part_of_the_model(self._model.snips_and_coefficients, self.initial_data,
                                                            function_data)
        range_for_initial = compute_range(
            (self._model.snips_and_coefficients[rsid] for rsid, _ in self.initial_data.items()), function_data)
        #print(self.initial_data.items().__len__())
        #print(self.imputed_data.items().__len__())
        #print(self.allele_freq_in_population.items().__len__())
        #print(self.rsids_absent_in_allele_freq_data.__len__())
        result_for_imputed_data = compute_part_of_the_model(self._model.snips_and_coefficients, self.imputed_data,
                                                            function_data)
        range_for_imputed = compute_range((
            self._model.snips_and_coefficients[rsid] for rsid, _ in self.imputed_data.items()), function_data)
        result_for_allele_freq_data = compute_part_of_the_model(self._model.snips_and_coefficients,
                                                                self.allele_freq_in_population,
                                                                function_data)
        range_for_allele_freq = compute_range((
            self._model.snips_and_coefficients[rsid] for rsid, _ in self.allele_freq_in_population.items()),
            function_data)
        result_for_not_found = function_data.neutral_number
        range_for_not_found = compute_range((
            self._model.snips_and_coefficients[rsid] for rsid in self.rsids_absent_in_allele_freq_data), function_data)

        range_for_model = sum([range_for_initial, range_for_imputed, range_for_allele_freq, range_for_not_found])
        result_for_model = function_data.function_to_apply_between_allele_pairs(
            [result_for_initial_data, result_for_imputed_data, result_for_allele_freq_data, result_for_not_found])
        category_boundaries = self._model.get_category_boundaries()
        # category_boundaries = transform_boundaries_if_needed(self._model, self._model.get_category_boundaries())

        return PolygenicRiskScoreResult(
            range_for_initial / range_for_model,
            range_for_imputed / range_for_model,
            range_for_allele_freq / range_for_model,
            range_for_not_found / range_for_model,
            result_for_initial_data,
            result_for_imputed_data,
            result_for_allele_freq_data,
            result_for_not_found,
            result_for_model,
            get_category(self._model.categories, result_for_model).scale_cat(result_for_model),
            get_category(self._model.categories, result_for_model).name,
            category_boundaries
        )


class PolygenicRiskScore(object):
    def __init__(self, categories: Iterable[QuantitativeCategory], snips_and_coefficients: Dict[str, ModelData],
                 model_type: str):
        super().__init__()
        implemented_model_types = {'beta', 'odds_ratio', 'prbs'}
        self.model_type = validate_model_type(model_type, implemented_model_types)
        self.categories = categories
        self.__validate_categories()
        self.snips_and_coefficients = snips_and_coefficients
        self.__data_already_read = False

    def get_rsids(self) -> Set[str]:
        return set(self.snips_and_coefficients.keys())

    def __validate_categories(self):
        validate_categories_names(self.categories)
        validate_categories_ranges(self.categories)

    def validate_allele_collection(self, allele_collection: Dict[str, SnpData], data_source: str):
        for rsid, snp_data in self.snips_and_coefficients.items():
            if rsid in allele_collection:
                alleles_in_data = set(allele_collection[rsid].alts)
                alleles_in_data.add(allele_collection[rsid].ref)
                validate_alleles(snp_data.effect_allele, alleles_in_data, rsid, data_source)

    def get_category_boundaries(self) -> Dict[str, Dict[str, float]]:
        return {category.name: {'from':category.from_, 'to': category.to} for category in self.categories}


def get_category(categories: List[QuantitativeCategory], num: float):
    for category in categories:
        if category.from_ <= num <= category.to:
            return category
    raise RuntimeError("No category matched")


def compute_boundaries(model_datas: Iterable[ModelData], function_data: FunctionData) -> Tuple[float]:
    positive_contributions = []
    negative_contributions = []
    for model_data in model_datas:
        extreme_results = [
            function_data.function_to_apply_on_allele_pair(2 * ['.'], model_data.effect_allele, model_data.coeff_value),
            function_data.function_to_apply_on_allele_pair(2 * [model_data.effect_allele], model_data.effect_allele,
                                                           model_data.coeff_value)
        ]
        extreme_results.sort()
        negative_contributions.append(extreme_results[0])
        positive_contributions.append(extreme_results[1])
    if negative_contributions:
        return function_data.function_to_apply_between_allele_pairs(
            negative_contributions), function_data.function_to_apply_between_allele_pairs(positive_contributions)
    else:
        return function_data.neutral_number, function_data.neutral_number


def compute_range(model_datas: Iterable[ModelData], function_data: FunctionData):
    #print("-----")
    #print(str(model_datas))
    #print(str(function_data))
    minimum, maximum = compute_boundaries(model_datas, function_data)
    return function_data.function_to_compute_ranges(maximum, minimum)


def validate_alleles(allele_in_model: str, alleles_in_data: Iterable[str], rsid: str, data_source: str):
    if allele_in_model not in alleles_in_data:
        print(str(allele_in_model) + " " + str(alleles_in_data))
        raise ConflictingAlleleBetweenDataAndModel(f'Conflicting alleles for {rsid} between model and {data_source}')


def validate_model_type(model_type: str, valid_model_types: Iterable[str]) -> str:
    small_letters_model_type = model_type.lower()
    if small_letters_model_type in valid_model_types:
        return small_letters_model_type
    raise NotImplementedError(f"Your model type: {small_letters_model_type} has not been implemented yet")


def compute_part_of_the_model(model_snips_and_coefficients: Dict[str, ModelData],
                              vcf_data: Union[Dict[str, SnpData], Dict[str, Tuple[str]]],
                              function_data: FunctionData) -> float:
    if not vcf_data:
        return function_data.neutral_number
    result = [None] * len(vcf_data)
    for i, (rsid, allele_data) in enumerate(vcf_data.items()):
        if isinstance(allele_data, SnpData):
            alleles = allele_data.genotype
        else:
            alleles = allele_data
        model_data = model_snips_and_coefficients[rsid]
        result[i] = function_data.function_to_apply_on_allele_pair(alleles, model_data.effect_allele,
                                                                   model_data.coeff_value)
    return function_data.function_to_apply_between_allele_pairs(result)


def rsids_data_from_data_source(rsids: Iterable[str], data_source: VcfAccessor, sample_name: str, is_imputed: bool = False) -> Dict[str, SnpData]:
    ret = {}
    for rsid in rsids:
        data = data_source.get_data_for_sample(sample_name, rsid, is_imputed)
        if data:
            ret[rsid] = data
    return ret


def get_allele_freq_in_poulation(snp_id: str, population_name: str,
                                 allele_frequency_accessor: AlleleFrequencyAccessor):
    logger.debug('Getting {} allele frequency from gnomad'.format(snp_id))
    return allele_frequency_accessor(snp_id, population_name)


def hardy_weinberg_diploid_frequencies(allele_frequencies: Dict[str, float]) -> Dict[Tuple[str], float]:
    assert len(allele_frequencies) == 2
    allele0 = list(allele_frequencies.keys())[0]
    allele1 = list(allele_frequencies.keys())[1]
    return {
        (allele0, allele0): allele_frequencies[allele0] ** 2,
        (allele1, allele1): allele_frequencies[allele1] ** 2,
        (allele0, allele1): allele_frequencies[allele1] * allele_frequencies[allele0] * 2
    }


def transform_boundaries_if_needed(model:PolygenicRiskScore, initial_boundaries:Dict[str, Dict[str, float]]):
    if model.model_type == 'beta':
        return initial_boundaries
    if model.model_type == 'odds_ratio':
       return change_boundaries(initial_boundaries)
    raise NotImplementedError(f'No known category boundaries transformation for model type {model.model_type}')


def change_boundaries(boundaries: Dict[str, Dict[str, float]]):
    ret_dict = {}
    for k, v in boundaries.items():
        ret_dict[k] = {}
        try:
            ret_dict[k]["from"] = math.log(v["from"])
        except ValueError:
            ret_dict[k]["from"] = np.NINF
        ret_dict[k]["to"] = math.log(v["to"])
    return ret_dict

if __name__ == '__main__':
    import glob, importlib, sys, os

    module_path = os.path.abspath(__file__).rsplit(os.path.sep, 6)[0]
    sys.path.insert(0, module_path)
    initial_paths = glob.glob('/data/test/large-data/vcf/mobigen_models_prepared/initial/*.vcf.gz')
    initial_accessor = VcfAccessor(initial_paths)
    imputed_paths = glob.glob('/data/test/large-data/vcf/mobigen_models_prepared/imputed/*.vcf.gz')
    imputed_accessor = VcfAccessor(imputed_paths)
    allele_accessor = AlleleFrequencyAccessor(
        allele_freq_json_path='/home/wojtek/PycharmProjects/mobigen/src/main/resources/allele_frequencies/gnomad.json')
    module = importlib.import_module('src/main/resources/test_models/testowy_nfe_model'.replace('/', '.'))
    model: PolygenicRiskScore = module.model
    data = Data(initial_accessor, imputed_accessor, allele_accessor, 'NA18500_9SR10670_D02.CEL', 'AF_nfe', model)
    print(data.compute_model())
    module2 = importlib.import_module('src/main/resources/test_models/testowy_covid_nfe_model'.replace('/', '.'))
    model2: PolygenicRiskScore = module2.model
    data2 = Data(initial_accessor, imputed_accessor, allele_accessor, 'NA18500_9SR10670_D02.CEL', 'AF_nfe', model2)
    print(data2.compute_model())
    module3 = importlib.import_module('src/main/resources/new_vitalleo_traits/covid_nfe_model'.replace('/', '.'))
    model3: PolygenicRiskScore = module3.model
    data3 = Data(initial_accessor, imputed_accessor, allele_accessor, 'NA18500_9SR10670_D02.CEL', 'AF_nfe', model3)
    print(data3.compute_model())



from unittest import TestCase
from polygenic import polygenic
from polygenic import polygenicmaker
from polygenic import vcfstat

import tabix
import os
import configparser

class PolygenicTest(TestCase):

    def test(self):
        self.assertEqual(1, 1)
    def testPolygenicCoreWithAf(self):
        polygenic.main([
            "--vcf", "test/resources/vcf/my.vcf.gz",
            "--sample", "yfyfy", 
            "--log-file", "/dev/null",
            "--model", "test/resources/model/scaled_eas_model.py", 
            "--population", "eas", 
            "--output-directory", "/tmp/polygenic",
            "--output-file-name", "bambala.json",
            "--af", "test/resources/vcf/af.vcf.gz"])
        self.assertEqual('1', '1')
    
    def testPolygenicForBiobankModel(self):
        polygenic.main([
            "--vcf", "test/resources/vcf/my.vcf.gz", 
            "--log_file", "/dev/null",
            "--model", "test/resources/model/biomarkers-30600-both_sexes-irnt.tsv.py", 
            "--population", "eas", 
            "--out_dir", "/tmp/",
            "--af", "test/resources/vcf/af.vcf.gz"])
        self.assertEqual('1', '1')

    def testPolygenicForGbeModel(self):
        polygenic.main([
            "--vcf", "test/resources/vcf/my.vcf.gz", 
            "--log_file", "/dev/null",
            "--model", "results/model/BIN1210.py", 
            "--population", "nfe", 
            "--out_dir", "/tmp/",
            "--af", "test/resources/vcf/af.vcf.gz"])
        self.assertEqual('1', '1')

class PolygenicMakerTest(TestCase):

    def testTabix(self):
        config = configparser.ConfigParser()
        config.read(os.path.dirname(__file__) + "/../polygenic/polygenic.cfg")
        url = config['urls']['hg19-rsids']
        tb = tabix.open(url)
        print(type(tb).__name__)
        records = tb.query("16", 1650945, 1650946)
        for record in records:
            print(record[:3])

    def testGbeIndex(self):
        polygenicmaker.main([
            "gbe-index",
            "--output", "results"
        ])

    def testGbeGet(self):
        polygenicmaker.main([
            "gbe-get",
            "--code", "BIN1210",
            "--output", "results"
        ])

    def testGbePrepareModel(self):
        polygenicmaker.main([
            "gbe-prepare-model",
            "--data", "results/BIN1210",
            "--af", "/home/marpiech/data/af.vcf.gz",
            "--output", "results/model"
        ])

    def testBiobankukIndex(self):
        polygenicmaker.main([
            "biobankuk-index",
            "--output", "results"
        ])
        self.assertEqual('1', '1')

    def testBiobankukGet(self):
        polygenicmaker.main([
            "biobankuk-get",
            "--index", "results/phenotype_manifest.tsv",
            "--phenocode", "30600",
            "--output", "results"
        ])
        self.assertEqual('1', '1')

    def testBiobankukBuildModel(self):
        polygenicmaker.main([
            "biobankuk-build-model",
            "--data", "results/biomarkers-30600-both_sexes-irnt.tsv",
            "--output", "results/model",
            "--anno", "results/full_variant_qc_metrics.txt",
            "--threshold", "1e-08"
        ])
        self.assertEqual('1', '1')

    def testPgsIndex(self):
        polygenicmaker.main([
            "pgs-index",
            "--output", "/tmp/polygenic"
        ])

    def testPgsGet(self):
        polygenicmaker.main([
            "pgs-get",
            "--code", "BIN1210",
            "--output", "results"
        ])

class VcfstatTest(TestCase):

    def testBaf(self):
        vcfstat.main([
            "baf",
            "--vcf", "/home/marpiech/data/clustered_204800980122_R01C02.vcf.gz",
            "--output-directory", "/tmp/baf"
        ])
        self.assertEqual('1', '1')

    def testZygosity(self):
        vcfstat.main([
            "zygosity",
            "--vcf", "/home/marpiech/data/clustered_204800980122_R01C02.vcf.gz",
            "--output-file", "/tmp/baf/stats.json"
        ])
        self.assertEqual('1', '1')

class Debug(TestCase):

    def testDebug(self):
        polygenic.main([
            "--vcf", "/tmp/clustered_204800980122_R01C02.vcf.gz",
            "--sample-name", "clustered_204800980122_R01C02",
            "--log-file", "/dev/null",
            "--output-name-appendix", "cancer",
            "--model", "/tmp/py/cancer1002.py",
            "--model", "/tmp/py/cancer1044.py", 
            "--population", "eas", 
            "--output-directory", "/tmp/polygenic/",
            "--af", "test/resources/vcf/af.vcf.gz"])
        self.assertEqual('1', '1')


#!/usr/bin/env python3
#
# -----------------------------------------------------------------------------
# Copyright (c) 2019, Battelle National Biodefense Institute.
#
# This file is part of rsidx (https://github.com/bioforensics/rsidx)
# and is licensed under the BSD license: see LICENSE.txt.
# -----------------------------------------------------------------------------

import rsidx
import sqlite3
from subprocess import Popen, PIPE
import sys


def trim_rsid(rsid):
    rsidstr = str(rsid)
    if rsidstr.startswith('rs'):
        rsidstr = rsidstr[2:]
    return rsidstr


def filter_by_rsid(instream, rsidlist, header=False):
    for line in instream:
        if line.startswith('#'):
            yield line
            continue
        chrom, coord, rsids, *values = line.split('\t')
        for rsid in rsids.split(';'):
            if rsid[2:] in rsidlist:
                yield line
                break


def search(rsidlist, dbconn, vcffile, header=False):
    c = dbconn.cursor()
    rsids = ', '.join(map(trim_rsid, rsidlist))
    query = 'SELECT DISTINCT chrom,coord FROM rsid_to_coord WHERE rsid IN ({:s})'.format(rsids)

    def fmt(row):
        return '{chr:s}:{coord:d}-{coord:d}'.format(chr=row[0], coord=row[1])
    coords = [fmt(result) for result in c.execute(query)]
    if len(coords) == 0:
        #print('[rsidx::search] WARNING: no rsID matches ' + str(rsidlist), file=sys.stderr)
        return
    coords.sort(
        key=lambda x: (x.split(':')[0], int(x.split(':')[1].split('-')[0]))
    )
    tabixcmd = ['tabix', vcffile]
    if header:
        tabixcmd.append('-h')
    tabixcmd.extend(coords)
    proc = Popen(tabixcmd, stdout=PIPE, universal_newlines=True)
    for line in filter_by_rsid(proc.stdout, rsids, header=header):
        yield line


def main(args):
    conn = sqlite3.connect(args.idx)
    with rsidx.open(args.out, 'w') as out:
        for line in search(args.rsid, conn, args.vcf, header=args.header):
            print(line, end='', file=out)
    conn.close()



import logging
from typing import List
from typing import Dict
from polygenic.lib.polygenic_exception import PolygenicException

logger = logging.getLogger('description_language.' + __name__)

class VcfRecord(object):
    def __init__(self, vcf_line:str, sample_names:List[str] = []):
        super().__init__()
        self.line = vcf_line

    def get_alt(self) -> List[str]:
        #try:
        return self.line.split("\t")[4].split(",")
        #except:
        #    raise PolygenicException("No alt allele in line: " + str(line))

    def get_ref(self) -> str:
        return self.line.split("\t")[3]

    def get_info(self) -> str:
        return self.line.split("\t")[7]

    def get_format(self) -> str:
        return self.line.split("\t")[8]

    def is_imputed(self) -> bool:
        return self.get_format().find("DS") != -1

    def get_info_field(self, name) -> str:
        for field in self.get_info().split(";"):
            field_name = field.split("=")[0]
            if field_name == name:
                return field.split("=")[1]

    def get_af_by_pop(self, population_name) -> Dict[str, float]:
        af = {}
        counter = 0
        sumfreq = 0
        for allele in self.get_alt():
            #print(population_name)
            #print(self.get_info())
            freq = float(self.get_info_field(population_name).split(",")[counter])
            sumfreq = sumfreq + freq
            af[allele] = freq
            counter = counter + 1
        af[self.get_ref()] = 1 - freq
        return af


import functools
import abc
import operator
import inspect
import logging
import re
import numpy as np
from typing import Callable
from typing import Iterable
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from polygenic.lib.allele_frequency_accessor import AlleleFrequencyAccessor

logger = logging.getLogger('description_language.' + __name__)

FUNCTION_DICT = {
    'add': sum,
    'multiply': functools.partial(functools.reduce, operator.mul)
}

FUNCTION_NAME_RE = re.compile('function (.*)\.<locals>')


class GenotypeGetter(object):

    def __init__(self, rsids: List) -> None:
        super().__init__()
        self.rsids = rsids

    def __call__(self, data):
        return {rsid: data[rsid] for rsid in self.rsids}


class Zygosity(abc.ABC):
    def __init__(self, allele):
        super().__init__()
        self.allele = allele

    @abc.abstractmethod
    def __eq__(self, other_genotype):
        if not (isinstance(other_genotype, list) or isinstance(other_genotype, tuple)):
            return False


class Homozygote(Zygosity):
    def __init__(self, allele: str):
        assert isinstance(allele, str)
        super().__init__(allele)

    def __eq__(self, other_genotype: Union[List, Tuple]):
        super().__eq__(other_genotype)
        other_genotype_set = set(other_genotype)
        return len(other_genotype_set) == 1 and self.allele in other_genotype_set


class Heterozygote(Zygosity):
    def __init__(self, alleles):
        assert isinstance(alleles, list) or isinstance(alleles, tuple) or isinstance(alleles, str)
        if isinstance(alleles, str):
            alleles = alleles.split('/')
        assert len(alleles) == 2
        super().__init__(alleles)

    def __eq__(self, other_genotype):
        super().__eq__(other_genotype)
        other_genotype_set = set(other_genotype)
        return len(other_genotype_set) == 2 and set(self.allele) == other_genotype_set


def at_least(number_of_conditions_to_met: int, conditions: List[Callable]):
    def f(data: Dict[str, Union[List, Tuple]], population_name: str,
          allele_frequency_accessor: AlleleFrequencyAccessor):
        conditions_met = 0
        for condition in conditions:
            if condition(data, population_name, allele_frequency_accessor):
                conditions_met += 1
            if conditions_met >= number_of_conditions_to_met:
                return True
        return False

    return f


def condition(rs_id, genotype, allele_frequency_accessor: AlleleFrequencyAccessor,
              use_population_allele_frequency=False) -> bool:
    def f(data: Dict[str, str], population_name: str):
        # data - słownik postaci rsid:'A/C'
        alleles = get_allele_frequency(rs_id, data, use_population_allele_frequency, population_name,
                                       allele_frequency_accessor)
        return genotype == alleles

    return f


def get_allele_frequency(rs_id: str, data: Dict[str, str], use_population_allele_frequency: bool, population_name: str,
                         allele_frequency_accessor: AlleleFrequencyAccessor):
    if use_population_allele_frequency:
        return get_alleles_when_missing_consider_population_freqs(rs_id, data, population_name,
                                                                  allele_frequency_accessor)
    else:
        return data[rs_id]


def get_value(coefficients_and_snps_to_be_considered: List[Callable], function_to_apply: str,
              function_dict: Dict[str, Callable] = FUNCTION_DICT):
    def f(data, population: str, allele_frequency_accessor: AlleleFrequencyAccessor):
        s = function_dict[function_to_apply](
            snp_function(data, population, allele_frequency_accessor) for snp_function in
            coefficients_and_snps_to_be_considered)
        return s

    return f


def get_category_and_value(categories: List, coefficients_and_snps_to_be_considered: List[Callable],
                           function_to_apply: str,
                           function_dict: Dict[str, Callable] = FUNCTION_DICT) -> Dict[str, Union[str, float]]:
    """
    Factory function that returns a function to compute category and numerical output for given coefficients and genotype
    :param categories: list of possible categories
    :param coefficients_and_snps_to_be_considered: list of callables to be applied to the genotype
    :param function_to_apply: string representation of function to be applied to the coefficients
    :param function_dict: translation between string representation of a function and the corresponding python callable
    :return: function to compute category and numerical output for given coefficients and genotype
    """

    def f(data, population: str, allele_frequency_accessor: AlleleFrequencyAccessor):
        s = function_dict[function_to_apply](
            snp_function(data, population, allele_frequency_accessor) for snp_function in
            coefficients_and_snps_to_be_considered)
        for category in categories:
            if category.from_ <= s <= category.to:
                returned_category_name = category.name
                break
        return {'categorical_output': returned_category_name, 'numerical_output': s}

    return f


def computation(coefficients_and_snps_to_be_considered: List[Callable], comparison_string: str, function_to_apply: str,
                function_dict: Dict[str, Callable] = FUNCTION_DICT):
    def f(data: Dict[str, str], population_name: str):
        return eval('{}{}'.format(function_dict[function_to_apply](
            snp_function(data, population_name) for snp_function in coefficients_and_snps_to_be_considered),
            comparison_string))

    return f


def odds_ratio(rs_id: str, effect_allele: str, odds_ratio_value: float) -> Callable:
    def func(data, population_name: str, allele_frequency_accessor: AlleleFrequencyAccessor):
        alleles = get_alleles_when_missing_consider_population_freqs(rs_id, data, population_name,
                                                                     allele_frequency_accessor)
        if len(alleles) < 1:
            logger.error('Improper genotype for {}: {}'.format(rs_id, alleles))
            return 1
        if alleles[0] == alleles[1] == effect_allele:
            return odds_ratio_value
        elif effect_allele in alleles:
            return 1
        else:
            return 1 / odds_ratio_value

    return func


class Priority(int):
    """
    Human-friendly representation of priority of a condition
    """

    def __new__(cls, priority: int):
        return super().__new__(cls, priority)

    def __gt__(self, other):
        if isinstance(other, self.__class__):
            return super().__gt__(self, other)
        else:
            raise TypeError('unorderable types: {} and {}'.format(type(self), type(other)))


class Choice(object):
    '''
    Model type which checks some conditions according to their priorities. When a condition is met, the part of the model for this condition is evaluated
    '''

    def __init__(self, priority_conditions_action_dict: Dict[Priority, Dict[str, Callable]]):
        """
        Initializes Choice instance
        :param priority_conditions_action_dict: dictionary of type {priority: {'when': function_that_tells_when_the_condition_is_evaluated, 'action': function_that_tells_how_to_evaluate_a_given_condition}}
        """
        super().__init__()
        self.priority_conditions_action_dict = priority_conditions_action_dict

    def __call__(self, data: Dict[str, str], allele_frequency_accessor: AlleleFrequencyAccessor, population):
        # data - a dictionary rsid:'A/C'
        for priority, condition_action_dict in sorted(self.priority_conditions_action_dict.items(), reverse=True):
            if condition_action_dict['when'](data, population, allele_frequency_accessor):
                return condition_action_dict['action'](data, population, allele_frequency_accessor)
        raise NoCategoryError("None od possible categories matched analysed data")

    def get_rsids_used_by_model(self) -> List[str]:
        # does not work for al_least function!
        callables = []
        for priority_value in self.priority_conditions_action_dict.values():
            if priority_value:
                # print('prio_val', priority_value['when'], inspect.getclosurevars(priority_value['when']).nonlocals)
                action_callables = self.__get_callable_from_priority_conditions_action_dict_val(priority_value,
                                                                                                'action')
                if action_callables:
                    callables.extend(action_callables)
                when_callables = self.__get_callable_from_priority_conditions_action_dict_val(priority_value, 'when')
                if when_callables:
                    callables.extend(when_callables)
        # print(callables)
        return [self.__get_rsid_from_callable(callable) for callable in callables]

    def category_boundaries(self, priority=Priority(0)) -> Dict[str, Dict[str, str]]:
        if priority != 0:
            raise NotImplementedError("Implmented only for priority = 0")
        categories = inspect.getclosurevars(self.priority_conditions_action_dict[priority]['action']).nonlocals[
            'categories']
        return {category.name: {'from': category.from_, 'to': category.to} for category in categories}

    def model_type(self) -> str:
        names_of_functions_used_to_compute_final_result = set()
        for _, v in self.priority_conditions_action_dict.items():
            function_name = FUNCTION_NAME_RE.search(str(v['action'])).group(1)
            if not function_name in ('get_value', 'get_category_and_value'):
                raise NotImplementedError('Not implemented for "{}" function'.format(function_name))
            for x in inspect.getclosurevars(v['action']).nonlocals['coefficients_and_snps_to_be_considered']:
                names_of_functions_used_to_compute_final_result.add(FUNCTION_NAME_RE.search(str(x)).group(1))
        assert len(names_of_functions_used_to_compute_final_result) == 1
        function_name = names_of_functions_used_to_compute_final_result.pop()
        return function_name

    @staticmethod
    def __get_callable_from_priority_conditions_action_dict_val(priority_conditions_action_dict_val: Dict, key: str) -> \
            Union[List[Callable], None]:
        try:
            return inspect.getclosurevars(
                priority_conditions_action_dict_val[key]).nonlocals['coefficients_and_snps_to_be_considered']
        except KeyError:
            pass

    @staticmethod
    def __get_rsid_from_callable(callale_with_rsid_in_closurevars: Callable) -> str:
        '''
        Method to extract rs_id from functions such as 'beta', 'odds_ratio', 'condition'
        :param callale_with_rsid_in_closurevars:
        :return:
        '''
        return inspect.getclosurevars(callale_with_rsid_in_closurevars).nonlocals['rs_id']


class NoCategoryError(RuntimeError):
    pass


def always_true(*args, **kwargs) -> bool:
    """
    Always returns true
    :param args:
    :param kwargs:
    :return: True
    """
    return True


def category(category_string: str) -> Callable:
    def f(_, __):
        return category_string

    return f


class QuantitativeCategory(object):
    def __init__(self, category_name, from_=-np.inf, to=np.inf):
        super().__init__()
        self.from_ = from_
        self.to = to
        self.name = category_name

    def intersects_with(self, other) -> None:
        if self.from_ == other.from_ or self.to == other.to:
            return True
        elif self.from_ < other.from_ < self.to:
            return True
        elif other.from_ < self.from_ < other.to:
            return True
        return False


class InvalidCategoriesError(RuntimeError):
    pass


class NoCategoryError(RuntimeError):
    """
    Raised  when no category could be returned.
    """
    pass


def get_alleles_when_missing_consider_population_freqs(rs_id: str, data: Dict[str, Union[List, Tuple]],
                                                       population_name: str,
                                                       allele_frequency_accessor: AlleleFrequencyAccessor):
    try:
        ret_val = data[rs_id]
        return ret_val
    except KeyError:
        logger.debug('{} was not found in input data'.format(rs_id))
        return max(hardy_weinberg_diploid_frequencies(
            get_allele_freq_in_poulation(rs_id, population_name, allele_frequency_accessor)).items(),
                   key=lambda x: x[1])[0]


def get_allele_freq_in_poulation(snp_id: str, population_name: str,
                                 allele_frequency_accessor: AlleleFrequencyAccessor):
    logger.debug('Getting {} allele frequency from gnomad'.format(snp_id))
    return allele_frequency_accessor(snp_id, population_name)


def hardy_weinberg_diploid_frequencies(allele_frequencies: Dict[str, float]) -> Dict[Tuple[str], float]:
    assert len(allele_frequencies) == 2
    allele0 = list(allele_frequencies.keys())[0]
    allele1 = list(allele_frequencies.keys())[1]
    return {
        (allele0, allele0): allele_frequencies[allele0] ** 2,
        (allele1, allele1): allele_frequencies[allele1] ** 2,
        (allele0, allele1): allele_frequencies[allele1] * allele_frequencies[allele0] * 2
    }


def beta(rs_id: str, effect_allele: str, beta_value: float) -> Callable:
    def func(data, population_name: str, allele_frequency_accessor: AlleleFrequencyAccessor):
        alleles = get_alleles_when_missing_consider_population_freqs(rs_id, data, population_name,
                                                                     allele_frequency_accessor)
        result = 0
        for allele in alleles:
            if allele == effect_allele:
                result += beta_value
        return result

    return func


def validate_categories(categories: Iterable[QuantitativeCategory]) -> None:
    validate_categories_names(categories)
    validate_categories_ranges(categories)


def validate_categories_names(categories: Iterable[QuantitativeCategory]):
    sorted_according_to_names = sorted(categories, key=lambda x: x.name)
    for i in range(len(sorted_according_to_names) - 1):
        if categories_have_same_names(sorted_according_to_names[i], sorted_according_to_names[i + 1]):
            raise InvalidCategoriesError(
                'Categories {} and {} have the same name'.format(sorted_according_to_names[i],
                                                                 sorted_according_to_names[i + 1]))


def categories_have_same_names(category1: QuantitativeCategory, category2: QuantitativeCategory):
    return category1.name == category2.name


def validate_categories_ranges(categories: Iterable[QuantitativeCategory]):
    for i in range(len(categories) - 1):
        for j in range(1, len(categories)):
            if categories[i].intersects_with(categories[j]):
                raise InvalidCategoriesError('Categories {} and {} intersect'.format(categories[i], categories[j]))


if __name__ == '__main__':
    print(Homozygote('A') == ['A', 'A'])
    het = Heterozygote('A/C')
    print(het.__eq__(['A', 'C']))
    print(het == ['C', 'A'])
    print(het == ['A', 'A'])
    print(het == ['T', 'T'])
    print(het == ['C'])



import argparse
import logging

import sys
import os
import urllib.request

# zygosity
import json

# baf
from pysam import VariantFile
import pandas as pandas
import plotly as plotly
import plotly.graph_objects as go
from plotly.subplots import make_subplots

def zygosity(args):
    parser = argparse.ArgumentParser(description='vcfstat zygosity draws baf images for chromosomes')  # todo dodać opis
    parser.add_argument('-v', '--vcf', type=str, required=True, help='input vcf.gz')
    parser.add_argument('-o', '--output-file', type=str, default='stats.json', help='output directory')
    parsed_args = parser.parse_args(args)
    chrom_sizes = pandas.read_csv(
        os.path.dirname(__file__) + '/resources/chromsizes/hg38.chrom.sizes', 
        delimiter='\t',
        index_col=0)
    chromosomes = list(chrom_sizes.index)
    vcf = VariantFile(parsed_args.vcf)

    out = []

    for sample_index in range(len(vcf.header.samples)):
        sample_name = vcf.header.samples[sample_index]
        out.append({})
        out[sample_index]["id"] = sample_name
        out[sample_index]["qc"] = {}
        for chrom in ['chrX', 'chrY', 'chrM']:
            size = chrom_sizes.loc[chrom]['size']
            df = pandas.DataFrame()
            count = 0
            het = 0
            alt = 0
            ref = 0

            for record in vcf.fetch(chrom, 1, size):
                if 'BAF' in record.format:
                    gt = record.samples.values()[sample_index]['GT']
                    if gt[0] == 0 and gt[1] == 0:
                        count = count + 1; ref = ref + 1
                    if gt[0] == 1 and gt[1] == 1:
                        count = count + 1; alt = alt + 1
                    if (gt[0] == 0 and gt[1] == 1) or (gt[0] == 1 and gt[1] == 0):
                        count = count + 1; het = het + 1
            out[sample_index]["qc"][chrom + "heterozygosity"] = {}
            out[sample_index]["qc"][chrom + "heterozygosity"]["value"] = het / count
            out[sample_index]["qc"][chrom + "heterozygosity"]["pass"] = "passed"
            out[sample_index]["qc"][chrom + "heterozygosity"]["plot"] = sample_name + "-" + chrom + ".jpg"

        out[sample_index]["qc"]["baflrr"] = {}
        out[sample_index]["qc"]["baflrr"]["value"] = "Click to view plots"
        out[sample_index]["qc"]["baflrr"]["pass"] = "passed"
        out[sample_index]["qc"]["baflrr"]["plot"] = sample_name + "-" + chrom + ".pdf"

    with open(parsed_args.output_file, 'w') as qcfile:
        json.dump(out, qcfile)
    
    return

def baf(args):
    parser = argparse.ArgumentParser(description='vcfstat baf draws baf images for chromosomes')  # todo dodać opis
    parser.add_argument('-v', '--vcf', type=str, required=True, help='input vcf.gz')
    parser.add_argument('-o', '--output-directory', type=str, default='', help='output directory')
    parsed_args = parser.parse_args(args)
    chrom_sizes = pandas.read_csv(
        os.path.dirname(__file__) + '/resources/chromsizes/hg38.chrom.sizes', 
        delimiter='\t',
        index_col=0)
    chromosomes = list(chrom_sizes.index)
    vcf = VariantFile(parsed_args.vcf)

    if not os.path.exists(parsed_args.output_directory):
        os.mkdir(parsed_args.output_directory)

    for sample_index in range(len(vcf.header.samples)):
        sample_name = vcf.header.samples[sample_index]
        for chrom in chromosomes:
            size = chrom_sizes.loc[chrom]['size']
            df = pandas.DataFrame()
            position = []
            baf = []
            lrr = []

            for record in vcf.fetch(chrom, 1, size):
                if 'BAF' in record.format:
                    position.append(record.pos)
                    baf.append(record.samples.values()[sample_index]['BAF'])
                    lrr.append(record.samples.values()[sample_index]['LRR'])
            df = pandas.DataFrame({'position': position, 'baf': baf, 'lrr': lrr})
        
            fig = make_subplots(
                rows=2, 
                cols=1,
                shared_xaxes=True,
                vertical_spacing=0.02)

            fig.add_trace(
                go.Scatter(
                    x = list(df['position']), 
                    y = list(df['baf']),
                    mode = "markers",
                    
                ),
                row=1, col=1
            )

            fig.add_trace(
                go.Scatter(
                    x = list(df['position']), 
                    y = list(df['lrr']),
                    mode = "markers"
                ),
                row=2, col=1
            )

            fig.update_traces(
                marker=dict(size=0.8)
            )

            fig.update_layout(
                height=600, 
                width=800, 
                title_text=sample_name + " " + chrom,
                template='simple_white',
                showlegend = False
                )

            fig.update_yaxes(
                row=1, col=1,
                title_text = "BAF",
                range = [-0.1, 1.1]
            )

            fig.update_yaxes(
                row=2, col=1,
                title_text = "LRR",
                range = [-1, 1]
            )

            fig.write_image(parsed_args.output_directory + "/" + sample_name + "-" + chrom + ".jpg")

    return

def main(args = sys.argv[1:]):
    #try:
        if args[0] == 'baf':
            baf(args[1:])
        elif args[0] == 'zygosity':
            zygosity(args[1:])
        else:
            raise Exception()
    # except Exception as e:
    #     print(e)
    #     print("ERROR " + str(e))
    #     print("""
    #     Program: vcfstat (calculates stats for vcfs)
    #     Contact: Marcin Piechota <piechota.marcin@gmail.com>

    #     Usage:   vcfstat <command> [options]

    #     Command: 
    #     baf                downloads pan biobankuk index of gwas results
    #     zygosity           downloads gwas results for given phenocode
    #     """)

if __name__ == '__main__':
    main(sys.argv[1:])



import setuptools

#! /usr/bin/python
# -*- coding: utf-8 -*-

from typing import List
from setuptools import setup, find_packages

PACKAGE_VERSION = '1.7.0'

def write_version_py(filename='polygenic/version.py'):
    cnt = """
# THIS FILE IS GENERATED FROM SETUP.PY
__version__ = '{}'
"""
    with open(filename, 'w') as f:
        f.write(cnt.format(PACKAGE_VERSION))


write_version_py()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="polygenic",
    version=PACKAGE_VERSION,
    author="Marcin Piechota, Wojciech Gałan",
    author_email="piechota@intelliseq.com",
    description="Polygenic score computation",
    #long_description=long_description,
    #long_description_content_type="text/reStructuredText",
    url="https://github.com/marpiech/polygenic",
    packages=setuptools.find_packages(include=['polygenic', 'polygenic.seqql']),
    package_data={'polygenic': ['*.cfg'], 'polygenic': ['resources/chromsizes/*.sizes']},
    license="Intelliseq dual licenses this package. For commercial use, please contact [contact @ intelliseq.com](mailto:contact@intelliseq.com). For non-commercial use, this license permits use of the software only by government agencies, schools, universities, non-profit organizations or individuals on projects that do not receive external funding other than government research grants and contracts. Any other use requires a commercial license. For the full license, please see [LICENSE.md](https://github.com/intelliseq/polygenic/blob/master/LICENSE.md), in this source repository.",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: Free for non-commercial use',
        'Programming Language :: Python :: 3',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    install_requires=[
        'numpy==1.20.3',
        'progressbar2==3.53.1',
        'python-utils==2.5.6',
        'pysam==0.16.0.1',
        'pytabix==0.1',
        'pandas==1.3.0',
        'plotly==5.1.0',
        'kaleido==0.2.1'
    ],
    python_requires='>=3.8',
    entry_points={
        'console_scripts': [
            'polygenic=polygenic.polygenic:main',
            'polygenicmaker=polygenic.polygenicmaker:main',
            'vcfstat=polygenic.vcfstat:main'
        ],
    },
    test_suite='nose.collector',
    tests_require=['nose>=1.0'],
    setup_requires=['nose>=1.0'],
)



import json
import logging
import importlib
from typing import Any
from typing import Dict
from typing import Iterable
from polygenic.seqql.score import PolygenicRiskScoreResult
from polygenic.lib.data_access.dto import ModelDescriptionInfo

logger = logging.getLogger('description_language.' + __name__)


# def create_res_representation_for_models(results: Dict[str, PolygenicRiskScoreResult],
#                                          descriptions: Dict[str, Iterable[Dict[str, Any]]],
#                                          populations: Dict[str, str]) -> Dict[str, Dict[str, Any]]:
#     ret = {}
#     for k, v in results.items():
#         ret[k] = create_res_representation_for_model(v, descriptions[k], populations[k])
#     return ret


def create_res_representation_for_model(result: PolygenicRiskScoreResult, model_description_info: ModelDescriptionInfo,
                                        population: str) -> Dict[str, Any]:
    ordered_dict_representation = result._asdict()
    ordered_dict_representation['trait_was_prepared_for_population'] = population
    module = importlib.import_module(model_description_info.model_fname.split('.')[0])
    descriptions = {}
    if module.description is None:
        print(module)
        for f_path in model_description_info.desc_paths:
            with open(f_path) as f:
                descriptions[f_path] = json.load(f)
        if model_description_info.plot_data_path:
            with open(model_description_info.plot_data_path) as f:
                ordered_dict_representation['plot_data'] = json.load(f)
        else:
            ordered_dict_representation['plot_data'] = None
    else:
        description = module.description
    ordered_dict_representation['descriptions'] = get_proper_values_depending_on_category_name(description, result.category)
    return dict(ordered_dict_representation)


def get_proper_values_depending_on_category_name(description:Dict[str, Any], category:str) -> Dict[str, Any]:
    out = {}
    for k, v in description.items():
        
        if k.endswith('_choice'):
            out[k] = v[category]
        else:
            out[k] = v
    return out



import os
import json
import logging
from typing import Dict
from typing import Union
from functools import partial
from polygenic.lib.allele_frequency_sqlite_dao import read_multiple_rs

logger = logging.getLogger('description_language.' + __name__)


class DataNotUsefulInModels(RuntimeError):
    pass


class AlleleFrequencyAccessor(object):

    def __init__(self, allele_freq_json_path: str = '', allele_freq_database_path: str = ''):
        assert not (allele_freq_json_path and allele_freq_database_path)
        if allele_freq_json_path:
            with open(allele_freq_json_path) as f:
                content = json.load(f)
            self.reading_method = partial(get_allele_freq_from_json, allele_freq_dict=content)
        if allele_freq_database_path:
            assert os.path.isfile(allele_freq_database_path)
            if not allele_freq_database_path.endswith('allele_frequency_db_gnomad_2.1.1.sqlite'):
                raise RuntimeError("implemented only for gnomad 2.1.1")
            self.reading_method = partial(get_allele_freq_from_db, path=allele_freq_database_path)

    def __call__(self, rsid: str, population: str):
        return self.reading_method(rsid, population)


def get_allele_freq_from_db(rsid: str, population_name: str, path: str = ''):
    snp_data = read_multiple_rs([rsid], path=path)[0]
    ref_allele = snp_data.ref
    alt_allele = snp_data.alt
    alt_allele_freq = snp_data.__dict__[population_name]
    if not snp_data.is_unique:
        logger.info(
            f'{rsid} is multiallelic but only two alleles are provided. Only {ref_allele} and {alt_allele} were considered')
    return {alt_allele: alt_allele_freq, ref_allele: 1 - alt_allele_freq}


def get_allele_freq_from_json(rsid: str, population_name: str, allele_freq_dict:Dict[str, Dict[str, Union[str, float, bool]]]):
    snp_data = allele_freq_dict[rsid]
    ref_allele = snp_data['ref']
    alt_allele = snp_data['alt']
    alt_allele_freq = snp_data[population_name]
    if not snp_data['is_biallelic']:
        logger.info(
            f'{rsid} is multiallelic but only two alleles are provided. Only {ref_allele} and {alt_allele} were considered')
    return {alt_allele: alt_allele_freq, ref_allele: 1 - alt_allele_freq}


if __name__ == '__main__':
    afa = AlleleFrequencyAccessor(
        allele_freq_database_path='/home/wojtek/PycharmProjects/mobigen/src/main/resources/databases/allele_frequency_db_gnomad_2.1.1.sqlite')
    print(afa('rs925966', 'AF_nfe'))
    afa2 = AlleleFrequencyAccessor(allele_freq_json_path='/home/wojtek/PycharmProjects/mobigen/src/main/resources/allele_frequencies/gnomad.json')
    print(afa2('rs925966', 'AF_asj'))



#!/usr/bin/env python3
#
# -----------------------------------------------------------------------------
# Copyright (c) 2019, Battelle National Biodefense Institute.
#
# This file is part of rsidx (https://github.com/bioforensics/rsidx)
# and is licensed under the BSD license: see LICENSE.txt.
# -----------------------------------------------------------------------------
# -*- coding: utf-8 -*-

from argparse import ArgumentParser, RawDescriptionHelpFormatter
import rsidx


def index_subparser(subparsers):
    cli = subparsers.add_parser('index')
    cli.add_argument(
        '-f', '--force', action='store_true', help='force overwrite of index '
        'file if it already exists'
    )
    cli.add_argument(
        '-c', '--cache-size', type=int, metavar='C', help='modify default '
        'sqlite3 cache size (in KiB)'
    )
    cli.add_argument(
        '-m', '--mmap-size', type=int, metavar='M', help='activate sqlite3 '
        'memory map mode and specify mmap_size (in bytes)'
    )
    cli.add_argument('vcf', help='sorted VCF file to index')
    cli.add_argument('idx', help='index file to create')


def search_subparser(subparsers):
    cli = subparsers.add_parser('search')
    cli.add_argument('--header', action='store_true', help='print VCF headers')
    cli.add_argument(
        '-o', '--out', metavar='FILE', help='write output to specified FILE; '
        'default is terminal (stdout)'
    )
    cli.add_argument('vcf', help='sorted and indexed VCF file')
    cli.add_argument('idx', help='rsidx index file')
    cli.add_argument('rsid', nargs='+', help='rsID(s) to search')


mains = {
    'index': rsidx.index.main,
    'search': rsidx.search.main,
}

subparser_funcs = {
    'index': index_subparser,
    'search': search_subparser,
}


def get_parser():
    # https://patorjk.com/software/taag/, "Standard" font
    bubbletext = r'''
≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠
           _     _
  _ __ ___(_) __| |_  __
 | '__/ __| |/ _` \ \/ /
 | |  \__ \ | (_| |>  <
 |_|  |___/_|\__,_/_/\_\

≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠≠
Invoke `rsidx index --help` and `rsidx search --help` for complete
usage instructions.
'''
    subcommandstr = ', '.join(sorted(list(mains.keys())))
    parser = ArgumentParser(
        description=bubbletext,
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser._positionals.title = 'Subcommands'
    parser._optionals.title = 'Global arguments'
    parser.add_argument('-v', '--version', action='version',
                        version='rsidx v{}'.format(rsidx.__version__))
    subparsers = parser.add_subparsers(dest='subcmd', metavar='subcmd',
                                       help=subcommandstr)
    for func in subparser_funcs.values():
        func(subparsers)
    return parser



import json
from typing import List
from typing import Dict


class WebsiteTextData(object):
    def __init__(self, title: str, accessed: str, link: str):
        super().__init__()
        self.title = title.strip('. ')
        self.accessed = accessed.strip('. ')
        self.links = link.strip('. ')


class ArticleLink(object):
    def __init__(self, link: str):
        super().__init__()
        self.link = link.strip('. ')


class References(object):
    def __init__(self, website_text_data_container: List[WebsiteTextData], article_links: List[ArticleLink]):
        super().__init__()
        self.website_text_data = website_text_data_container
        self.article_links = article_links

    def __str__(self):
        return json.dumps(self.to_dict())

    def to_dict(self) -> Dict[str, List]:
        return {k: [sub_val.__dict__ for sub_val in v] for k, v in self.__dict__.items()}


if __name__ == '__main__':
    print(WebsiteTextData('bla', 'acc', 'l1'))
